from pathlib import Path
import itertools
import csv
import numpy
import math
from scipy import signal

import digichem.log
from digichem.parse.base import File_parser_abc
from digichem.exception import Digichem_exception
from digichem.misc.io import checksum

# Hidden imports.
#import cclib.io


class Cclib_parser(File_parser_abc):
    """
    ABC for parsers that use cclib to do most of their work for them.
    """
    
    # A dictionary of recognised auxiliary file types.
    INPUT_FILE_TYPES = {}
    
    def __init__(self, *log_files, options, **auxiliary_files):
        """
        Top level constructor for calculation parsers.
        
        :param log_files: A list of output file to analyse/parse. The first log_file given will be used for naming purposes.
        :param auxiliary_files: A dictionary of auxiliary input files related to the calculation.
        """
        # Also save our aux files, stripping None.
        self.auxiliary_files = {name: aux_file for name,aux_file in auxiliary_files.items() if aux_file is not None}

        # TODO: Does this belong here?
        # Also have a look for a profile.csv file that we can us for performance metrics.
        self.profile_file = Path(log_files[0].parent, "../Logs/profile.csv")
        
        super().__init__(*log_files, options = options)
        
    @classmethod
    def from_logs(self, *log_files, hints = None, options, **kwargs):
        """
        Intelligent constructor that will attempt to guess the location of files from a given log file(s).
        
        :param given_log_files: Output file(s) to parse or a directory of output files to parse.
        """
        # Have a look for aux. files.
        auxiliary_files = {}

        basename = log_files[0].name if len(log_files) > 0 else ""
        
        for hint in itertools.chain(log_files, hints if hints is not None else []):
            auxiliary_files.update(self.find_auxiliary_files(hint, basename))
            
        # Finally, update our auxiliary_files with kwargs, so any user specified aux files take precedence.
        auxiliary_files.update(kwargs)
        
        return self(*log_files, options = options, **auxiliary_files)
    
    @classmethod
    def find_auxiliary_files(self, hint, basename):
        """
        Find auxiliary files from a given hint.
        
        :param hint: A path to a file to use as a hint to find additional files.
        :returns: A dictionary of found aux files.
        """
        hint = Path(hint)
        
        # Now have a look for aux. input files, which are defined by each parser's INPUT_FILE_TYPES
        auxiliary_files = {}
        for file_type in self.INPUT_FILE_TYPES:
            for extension in file_type.extensions:
                if hint.is_dir():
                    # Peak inside.
                    if Path(hint, basename).with_suffix(extension).exists():
                        auxiliary_files[self.INPUT_FILE_TYPES[file_type]] = Path(hint, basename).with_suffix(extension)

                if hint.with_suffix(extension).exists():
                    auxiliary_files[self.INPUT_FILE_TYPES[file_type]] = hint.with_suffix(extension)
        
        return auxiliary_files
    
    def _parse(self):
        """
        Extract results from our output files.
        
        This default implementation will parse the log file with cclib and save the result to self.data.
        """
        import cclib.io
        
        # We start by using cclib to get most of the data we need.
        
        # Output a message (because this is slow).
        digichem.log.get_logger().info("Parsing calculation result '{}'".format(self.description))
        
        # Use cclib to open our log files.
        # ccread will accept a list of log files to read, but will sometimes choke if the list contains only one entry,
        # in which case we give it only one file name.
        # ccread will also choke if we give it pathlib Paths.
        file_paths = [str(log_file) for log_file in self.log_file_paths]
        
        # Get data from cclib.
        self.data = cclib.io.ccread(file_paths if len(file_paths) > 1 else file_paths[0])
        if self.data is None:
            raise Digichem_exception("Could not parse any data at all!")
        
        # Get a unique ID (checksum) from the given log files.
        # First, order the list of filenames so we also process in the same order.
        # We do this because not all parsers define a custom sort.
        
        file_paths.sort()
        self.data._id = checksum(*file_paths, hash_func = "sha1")
        
        # Do some setup.
        self.pre_parse()
        
        # Do our own parsing (if any).
        for log_file_path in self.log_file_paths:
            with open(log_file_path, "rt") as log_file:
                for line in log_file:
                    self.parse_output_line(log_file, line)

        # Add profiling data.
        try:
            self.parse_profile_file()
        
        except Exception:
            if self.profile_file.exists():
                digichem.log.get_logger().warning("Could not parse profile.csv file; profiling data will be unavailable", exc_info=True)
            
            else:
                pass

    def parse_profile_file(self):
        """
        """
        # Real calculations can end up with millions of rows, which is far too much data to handle.
        # We will need to downsample if we have too many data points.
        # First work out how many rows there are.
        try:
            with open(self.profile_file, "rb") as profile_file:
                lines = sum(1 for _ in profile_file) -1 # Remove 1 line for the header.
        
        except FileNotFoundError:
            # This is ok
            return

        if lines < 2:
            return

        max_lines  = self.options.parse['profiling_rows']
        factor = math.ceil(lines / max_lines)
        
        with open(self.profile_file) as profile_file:
            reader = csv.reader(profile_file)

            # Get the header.
            headers = next(reader)

            # Check headers match.
            if (headers[0] == "Duration / s" and 
               headers[1] == "Memory Used (Real) / bytes" and
               headers[2] == "Memory Used (Real) / %" and
               headers[3] == "Memory Available  (Real) / bytes" and
               headers[4] == "Memory Available (Real) / %" and
               headers[9] == "CPU Usage / %" and
               headers[15] == "Output Directory Available / bytes" and
               headers[17] == "Scratch Directory Used / bytes" and
               headers[18] == "Scratch Directory Available / bytes"
            ):
                column_map = {
                    "duration": 0,
                    "memory_used": 1,
                    "memory_used_percent": 2,
                    "memory_available": 3,
                    "memory_available_percent": 4,
                    "cpu_used": 9,
                    "output_available": 15,
                    "scratch_used": 17,
                    "scratch_available": 18
                }
            
            elif (headers[0] == "Duration / s" and 
               headers[1] == "Memory Used (Real) / bytes" and
               headers[2] == "Memory Used (Real) / %" and
               headers[3] == "Memory Available  (Real) / bytes" and
               headers[4] == "Memory Available (Real) / %" and
               headers[9] == "CPU Usage / %" and
               headers[15] == "Output Directory Available / bytes" and
               headers[17] == "Scratch Directory Available / bytes"
            ):
                column_map = {
                    "duration": 0,
                    "memory_used": 1,
                    "memory_used_percent": 2,
                    "memory_available": 3,
                    "memory_available_percent": 4,
                    "cpu_used": 9,
                    "output_available": 15,
                    "scratch_available": 17
                }
            
            else:
                raise Digichem_exception("wrong headers found in profile.csv file")
            
            # Then the body.
            # TODO: Reading the entire file is not ideal...
            data = numpy.genfromtxt(
                profile_file,
                delimiter=',',
                # TODO: use something better.
                filling_values = "0"
            )

        # We'll keep:
        # - duration
        # - memory used
        # - memory used %
        # - memory available
        # - memory available %
        # - cpu used
        # - output space
        # - scratch space
        new_data = numpy.zeros((math.ceil(lines / factor), len(column_map)))

        # Now decimate.
        for i, k in enumerate(column_map.values()):
            if factor > 1:
                new_data[:, i] = signal.decimate(data[:, k], factor)
            else:
                new_data[:, i] = data[:, k]
        
        
        self.data.metadata['performance'] = {
            key: new_data[:, index] for index, key in enumerate(column_map)
        }
        
        
    def parse_output_line(self, log_file, line):
        """
        Perform custom line-by-line parsing of an output file.
        
        This method will be called for each line of each log-file given to the parser (although be aware that some implementations may skip some lines during parsing),
        and it allows for data not supported by cclib to be extracted. It is program specific.
        """
        # Do nothing.
        
    def pre_parse(self):
        """
        Perform any setup before line-by-line parsing.
        """
        # Do nothing.
        
    @classmethod
    def au_to_debye(self, au):
        """
        Convert a dipole moment in au to debye.
        """
        return au * 2.541746473
    
    @classmethod
    def bohr_to_angstrom(self, bohr_distance):
        """
        Convert a length in bohr to angstrom.
        """
        return bohr_distance * 0.529177210903


# Functions and classes for determining file types etc.
from pathlib import Path
from itertools import accumulate

from digichem.exception.base import Unknown_file_type_exception

class File_type():
    """
    Class for representing a known file type.
    """
    
    def __init__(self, name, family = None, extensions = None, short_code = None):
        """
        Constructor for File_type objects.
        
        :param name: The name of the file (eg, "JPEG").
        :param family: The name of the family/group that the file type belongs to (eg, "image").
        :param extensions: An iterable of known file extensions (this is case-insensitive) (eg, [".jpg", ".jpeg"]).
        """
        self.name = name
        self.family = family
        self.extensions = [extension.lower() for extension in extensions] if extensions is not None else []
        
        if short_code is None and len(self.extensions) > 0:
            # Take one of our extensions (without the dot).
            self.short_code = self.extensions[0][1:]
            
        else:
            self.short_code = short_code
    
    def check(self, file_path):
        """
        Determine whether a given file is of this file type.
        
        :param file_path: A string-like path to the file to check.
        :return: True if the file is of this type, false otherwise.
        """
        # Get a Path object.
        file_path = Path(file_path)
        
        # Check to see if the file extension matches one of our file extensions.
        # We recurse through all the given file's extensions in case one of our extensions contains multiple parts (".tar.gz" for example).
        for extension in accumulate(reversed(file_path.suffixes), lambda a, b: b+a):
            if extension.lower() in self.extensions:
                return True
        
        # No match found.
        return False
    
    @property
    def file_type(self):
        """
        Get a string uniquely identifying this file type.
        """
        if self.family is not None:
            return "{}/{}".format(self.family, self.name)
        else:
            return self.name
        
    def __str__(self):
        """
        Stringify this file type.
        """
        return self.file_type
    
    
# Known file types.
log_file = File_type("log", "general", [".log"])
gaussian_chk_file = File_type("checkpoint", "gaussian", [".chk"])
gaussian_NTO_chk_file = File_type("NTO-checkpoint", "gaussian", [".chk"])
gaussian_fchk_file = File_type("formatted-checkpoint", "gaussian", [".fchk"])
gaussian_rwf_file = File_type("read-write", "gaussian", [".rwf"])
gaussian_cube_file = File_type("cube", "gaussian", [".cub", ".cube"])

orca_gbw_file = File_type("gbw", "orca", [".gbw"])
orca_density_file = File_type("density", "orca", [".densities"])
orca_density_info_file = File_type("density-info", "orca", [".densitiesinfo"])

# A list of all our known types.
known_types = [log_file, gaussian_chk_file, gaussian_fchk_file, gaussian_rwf_file, gaussian_cube_file, orca_gbw_file]

# TODO: This seems to be unused?
def get_file_type(file_path):
    """
    Get the type of a file from a list of known file types.
    
    :raises Unknown_file_type_exception: If the given file is of an unknown type.
    :param file_path: String-like path to the file to check.
    :return: A File_type object.
    """
    # Iterate through our list of known file types.
    for file_type in known_types:
        if file_type.check(file_path):
            return file_type
        
    # Unknown file type.
    raise Unknown_file_type_exception(file_path)




# General imports.
import hashlib

# Digichem imports.
from digichem.result.result import Result_set
from digichem.result.metadata import Metadata
from digichem.result.atom import Atom_list
from digichem.result.ground_state import Ground_state
from digichem.result.excited_state import Excited_state_list
from digichem.result.emission import Relaxed_excited_state
from digichem.result.orbital import Molecular_orbital_list
from digichem.result.alignment.base import Alignment, Minimal


class Merged(Result_set):
    """
    A type of result set that represents one or more separate calculations merged into one result set.
    """
    
    def __init__(self, results, *args, vertical_emission = None, adiabatic_emission = None, **kwargs):
        """
        Constructor for Merged result sets.
        :param vertical_emission: An optional dictionary of Relaxed_excited_state objects representing vertical emission energies (one for each multiplicity).
        :param adiabatic_emission: An optional dictionary of Relaxed_excited_state objects representing vertical adiabatic energies (one for each multiplicity).
        """
        super().__init__(*args, **kwargs)
        self.results = results
        self.emission.vertical = vertical_emission if vertical_emission is not None else {}
        self.emission.adiabatic = adiabatic_emission if adiabatic_emission is not None else {}
            
    @classmethod
    def from_results(self, *results, options):
        """
        Create a Merged result set object from a number of result sets.
        
        :param *results: List of result sets to merge.
        :param options: A Digichem config nested dictionary containing options to control parsing.
        """
        # First, get a merged metadata object.
        metadatas = [result.metadata for result in results]
        merged_metadata = Metadata.merge(*metadatas)

        # 'Merge' our atoms.
        raw_atoms = Atom_list.merge(*[result.raw_atoms for result in results], charge = merged_metadata.charge)
        # And alignment.
        alignment_class = Alignment.from_class_handle(options['alignment']) if options['alignment'] is not None else Minimal
        atoms = alignment_class(raw_atoms, charge = merged_metadata.charge)
        
        # Use special method for merging orbitals.
        orbitals, beta_orbitals = Molecular_orbital_list.merge_orbitals([result.orbitals for result in results], [result.beta_orbitals for result in results])
        
        # Merge NMR (which depends on atoms and options).
        nmr = type(getattr(results[0], "nmr")).merge(*[getattr(result, "nmr") for result in results], atoms = atoms, options = options)
        
        # Merge remaining attributes.
        attrs = {}
        for attr in ["energies", "pdm", "excited_states", "vibrations", "soc"]:
            attrs[attr] = type(getattr(results[0], attr)).merge(*[getattr(result, attr) for result in results])
        
        # Get a new ground state.
        ground_state = Ground_state.from_energies(merged_metadata.charge, merged_metadata.multiplicity, attrs['energies'])
        
        # Build new list of energy states.
        energy_states = Excited_state_list()
        energy_states.append(ground_state)
        energy_states.extend(attrs['excited_states'])
        
        # Get a new ID for the merged object.
        # This is generated by combining the hashes of each of the contributing results.
        hasher = hashlib.sha1()
        
        for result in results:
            if result._id:
                byte_hash = bytes.fromhex(result._id)
                
            hasher.update(byte_hash)
            
        _id = hasher.hexdigest()
        
        merged_results =  self(
            _id = _id,
            metadata = merged_metadata,
            results = results,
            ground_state = ground_state,
            raw_atoms = raw_atoms,
            atoms = atoms,
            energy_states = energy_states,
            orbitals = orbitals,
            beta_orbitals = beta_orbitals,
            nmr = nmr,
            **attrs
            )
        
        # Try and guess emission.
        merged_results.emission.vertical, merged_results.emission.adiabatic = Relaxed_excited_state.guess_from_results(*results)
        
        # Done.
        return merged_results



import pytest
from pathlib import Path
import yaml
import copy

from digichem.test.util import digichem_options, data_directory
from digichem.config import Auto_type, get_config
from digichem.config.parse import Config_parser

def test_config_save(digichem_options, tmpdir):
    """Can we save config options to file?"""
    # First, check an exception is thrown if we try and save to a non-existent dir:
    # with pytest.raises(Exception):
    #     digichem_options.save(Path(tmpdir, "foobar", "conf.yaml"))

    # Now save properly.
    save_file = Path(tmpdir, "conf.yaml")
    digichem_options.save(save_file)

    # Open the file and read it back.
    with open(save_file) as file:
        data = yaml.safe_load(file)

    # Check the data is the same.
    assert digichem_options.dump() == data

@pytest.mark.parametrize("var, typeis", [
    ["1", int],
    ["1.1", float],
    ["one", str]
], ids=["int", "float", "str"])
def test_autotype(var, typeis):
    """Does autotype work?"""
    assert type(Auto_type(var)) is typeis

def test_string_parser():
    """Can we parse config options from a string."""
    Config_parser("foo: bar").load() == {"foo": "bar"}

def test_bad_string_parser():
    """Can we not parse config options from a malformed string."""
    with pytest.raises(Exception):
        Config_parser("foo: {bar").load()

def test_del_option(digichem_options):
    """Can we reset an option to its default by deleting it?"""
    options = copy.deepcopy(digichem_options)
    # First, set a value.
    options.alignment = "FAP"

    # Delete and check.
    del options.alignment
    assert options.alignment == "MIN"

def test_config_merge():
    """Can we set additional options from a string?"""
    conf = get_config(
        clear_cache = True,
        extra_config_strings = ["alignment: FAP"]
    )

    assert conf.alignment == "FAP"
    get_config(clear_cache = True)

def test_config_file_merge(tmpdir):
    """Can we set additional options from another file?"""
    with open(Path(tmpdir, "conf.yaml"), "w") as conf_file:
        yaml.dump({'alignment': 'FAP'}, conf_file)

    conf = get_config(
        clear_cache = True,
        extra_config_files = [Path(tmpdir, "conf.yaml")]
    )

    assert conf.alignment == "FAP"

    # Reset the config object.
    get_config(clear_cache = True)


import argparse


class Extend_action(argparse.Action):
    """
    An argparse action class to replace the "extend" action for versions of python (3.6 etc) that do not support it natively.
    """

    def __call__(self, parser, namespace, values, option_string = None):
        """
        """
        try:
            attr = getattr(namespace, self.dest)
            
        except AttributeError:
            attr = []
            setattr(namespace, self.dest, attr)
            
        attr.extend(values)

# TODO: Unused.
class List_grouper(argparse.Action):
    """
    Custom action class that groups lists together so we know in what order they were specified.
    """    
    
    def __init__(self, option_strings, *args, **kwargs):
        """
        """
        argparse.Action.__init__(self, option_strings, *args, **kwargs)
        # We'll give ourself a name so we also group no matter which of our option_strings is used.
        self.name = [name for name in option_strings if name[:2] == "--"]
    
    def __call__(self, parser, namespace, values, option_string=None):
        """
        """
        grouped_list = {
            'group': self,
            'values': values
        }
        try:
            getattr(namespace, self.dest).append(grouped_list)
        except AttributeError:
            setattr(namespace, self.dest, [grouped_list])


# General imports
import math

# Digichem imports
from digichem.result import Result_object
from digichem.result.angle import Angle


class Dipole_moment_ABC(Result_object):
    """
    ABC that represents a dipole moment.
    
    See inheriting classes for concrete implementations (depending on whether they are permanent or transition, and electric or magnetic).
    """
    
    # A warning issued when attempting to merge non-equivalent objects
    MERGE_WARNING = "Attempting to merge list of DPMs that are not identical; non-equivalent DPMs will be ignored"
    
    @property
    def magnitude(self):
        """
        The dipole moment total (root-sum-square of the dipole moment vector).
        
        :deprecated: 'magnitude' is a somewhat misleading name for this attribute, use 'total' instead.
        """
        return self.total
    
    @property
    def total(self):
        """
        The dipole moment total (root-sum-square of the dipole moment vector).
        """
        return math.sqrt( self.vector_coords[0] ** 2 + self.vector_coords[1] ** 2 + self.vector_coords[2] ** 2)
    
    def __init__(self, origin_coords, vector_coords, atoms = None):
        """
        Constructor for Dipole_moment objects.
        
        :param origin_coords: Tuple of (x, y, z) coordinates of the origin of this dipole moment.
        :param vector_coords: Tuple of (x, y, z) coordinates of this dipole  moment.
        :param atoms: Optional atoms atom list object which is used to calculate addition data about this dipole moment. 
        """
        super().__init__()
        # The start of our vector, normally (0,0,0).
        self.origin_coords = origin_coords
        # The end of our vector.
        self.vector_coords = vector_coords
        
        # Save our atoms object.
        self.atoms = atoms if atoms is not None else []
        
        # Save a name describing which dipole we are (permanent vs transition etc).
        self.dipole_type = "permanent"
    
    @property
    def name(self):
        """
        Name that describes this dipole moment.
        """
        return "PDM"
    
    @property
    def origin_coords(self):
        """
        The origin coords of this vector as a tuple of (x, y, z). origin_coords is automatically realigned by the atoms alignment object of this dipole, use _origin_coords if you do not want this behaviour.
        """
        if len(self.atoms) > 0:
            return self.atoms.apply_transformation(self._origin_coords)
        else:
            return self._origin_coords
        
    @origin_coords.setter
    def origin_coords(self, value):
        """
        Set the origin coords of this dipole.
        """
        self._origin_coords = value
        
    @property
    def vector_coords(self):
        """
        The ending coords of this vector as a tuple of (x, y, z). vector_coords is automatically realigned by the atoms alignment object of this dipole, use _vector_coords if you do not want this behaviour.
        """
        if len(self.atoms) > 0:
            return self.atoms.apply_transformation(self._vector_coords)
        else:
            return self._vector_coords
        
    @vector_coords.setter
    def vector_coords(self, value):
        """
        Set the ending coords of this dipole.
        """
        self._vector_coords = value
        
    def __float__(self):
        """
        Floatify this Dipole_moment.
        """
        return self.magnitude
    
    def __eq__(self, other):
        """
        Is this dipole moment equal to another?
        
        A dipole moment is considered equivalent only if all coordinates match.
        """
        return self.origin_coords == other.origin_coords and self.vector_coords == other.vector_coords
    
    @property
    def X_axis_angle(self):
        """
        The angle between this dipole moment and the X axis of the atom set of the molecule.
        """
        if self.total == 0:
            return Angle(0)
        elif len(self.atoms) > 0:
            return Angle(math.fabs(self.atoms.get_X_axis_angle(self.origin_coords, self.vector_coords)), "rad")
        else:
            return None
        
    @property
    def XY_plane_angle(self):
        """
        The angle between this dipole moment and the XY plane of the atom set of the molecule.
        """
        if self.total == 0:
            return Angle(0)
        elif len(self.atoms) > 0:
            return Angle(math.fabs(self.atoms.get_XY_plane_angle(self.origin_coords, self.vector_coords)), "rad")
        else:
            return None
    
    @property
    def gaussian_cgs_vector(self):
        """
        The vector of this dipole moment in Gaussian-cgs units.
        
        Note that the sign of the vector will be reversed in Gaussian-cgs units.
        See J. Phys. Chem. Lett. 2021, 21, 686-695.
        """
        return tuple(self.to_gaussian_cgs(coord) for coord in self.vector_coords)
    
    @property
    def gaussian_cgs(self):
        """
        The total of this dipole moment in Gaussian-cgs units.
        
        Note that the sign of the vector will be reversed in Gaussian-cgs units.
        See J. Phys. Chem. Lett. 2021, 21, 686-695.
        """
        cgs_vector = self.gaussian_cgs_vector
        return math.sqrt( cgs_vector[0] **2 + cgs_vector[1] **2 + cgs_vector[2] **2 )
    
    def angle(self, other, cgs = True):
        """
        Find the angle between this dipole moment and another.
        
        :param cgs: Whether to find the angle in Gaussian-cgs units or standard units (the direction of magnetic dipole moments is reversed in the former).
        """
        return Angle(math.acos(self.cos_angle(other, cgs)))
        
    def cos_angle(self, other, cgs = True):
        """
        Find the cosine of the angle between this dipole moment and another.
        
        :param cgs: Whether to find the angle in Gaussian-cgs units or standard units (the direction of magnetic dipole moments is reversed in the former).
        """
        if cgs:
            vector_a = self.gaussian_cgs_vector
            vector_b = other.gaussian_cgs_vector
        
        else:
            vector_a = self.vector_coords
            vector_b = other.vector_coords
            
        vector_a_total = math.sqrt( vector_a[0] **2 + vector_a[1] **2 + vector_a[2] **2 )
        vector_b_total = math.sqrt( vector_b[0] **2 + vector_b[1] **2 + vector_b[2] **2 )
        
        try:
            return (vector_a[0] * vector_b[0] + vector_a[1] * vector_b[1] + vector_a[2] * vector_b[2]) / (vector_a_total * vector_b_total)
        except (FloatingPointError, ZeroDivisionError):
            return 0
        
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "total": {
                "value": self.total,
                "units": self.units
            },
            "origin": {
                "x": {
                    "value": float(self.origin_coords[0]),
                    "units": "Å",
                },
                "y": {
                    "value": float(self.origin_coords[1]),
                    "units": "Å",
                },
                "z": {
                    "value": float(self.origin_coords[2]),
                    "units": "Å",
                }
            },
            "vector": {
                "x": {
                    "value": float(self.vector_coords[0]),
                    "units": self.units
                },
                "y": {
                    "value": float(self.vector_coords[1]),
                    "units": self.units
                },
                "z": {
                    "value": float(self.vector_coords[2]),
                    "units": self.units
                },
            },
            "x-angle": {
                "value": self.X_axis_angle.angle,
                "units": self.X_axis_angle.units
            },
            "xy-angle": {
                "value": self.XY_plane_angle.angle,
                "units": self.XY_plane_angle.units
            }
        }


class Electric_dipole_moment_mixin():
    """
    Mixin class for electric dipole moments.
    """
    
    @property
    def units(self):
        """
        The units of this dipole moment object.
        """
        # Debye
        return "D"
    
    @property
    def coulomb_meters(self):
        """
        The total of this dipole moment in Coulomb Meters.
        """
        return float(self) * 3.335640952e-30
    
    @classmethod
    def D_to_au(self, value):
        """
        Convert an electric dipole moment in D to a.u.
        """
        return value / 2.541746473
    
    @classmethod
    def to_gaussian_cgs(self, value):
        """
        Convert an electric dipole moment in D to Gaussian-cgs units.
        See J. Phys. Chem. Lett. 2021, 21, 686-695.
        """
        # cgs = au * e * a0
        return self.D_to_au(value) * 4.80320425e-10 * 5.29177210903e-9


class Magnetic_dipole_moment_mixin():
    """
    Mixin class for magnetic dipole moments.
    """
    
    @property
    def units(self):
        """
        The units of this dipole moment object.
        """
        return "a.u."
    
    @property
    def bohr_magnetons(self):
        """
        The total of this magnetic dipole moment in bohr magnetons.
        """
        return self.total * 2
    
    @property
    def ampere_square_meters(self):
        """
        The total of this magnetic dipole moment in SI units.
        """
        return self.total * 1.8548e-23
    
    @classmethod
    def to_gaussian_cgs(self, value):
        """
        Convert a magnetic dipole moment in a.u. to Gaussian-cgs units.
        See J. Phys. Chem. Lett. 2021, 21, 686-695.
        """
        return value * -9.2740100783e-21


class Dipole_moment(Dipole_moment_ABC, Electric_dipole_moment_mixin):
    """
    Class that represents the (permanent) dipole moment of a molecule.
    """
        
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self((data['origin']['x']['value'], data['origin']['y']['value'], data['origin']['z']['value']), (data['vector']['x']['value'], data['vector']['y']['value'], data['vector']['z']['value']), atoms = result_set.atoms)
    
    @classmethod
    def from_parser(self, parser):
        """
        Construct a Dipole_moment object from an output file parser.
        
        :param parser: An output file parser.
        :result: A single Dipole_moment object, or None if no dipole information is available.
        """
        try:
            return self(parser.data.moments[0], parser.data.moments[1], parser.results.atoms)
        except AttributeError:
            return None
    


# General imports.
import yaml
from pathlib import Path
import packaging.version
import dill
import periodictable

# Digichem imports.
from digichem.exception.base import Digichem_exception
from digichem.input import Gaussian_input_parser
from digichem.file.prattle import Openprattle_converter, Oprattle_formats
import digichem.log
from digichem.input.base import Input_file
from digichem.parse.util import parse_calculation, open_for_parsing
import digichem.config
from digichem.result.atom import Molecule_mixin

# Custom formats to allow literal strings in yaml output.
# Adapted from https://stackoverflow.com/questions/6432605/any-yaml-libraries-in-python-that-support-dumping-of-long-strings-as-block-liter

# Custom 'str' class that will be dumped literally.
class literal_str(str): pass
# The dumper (which recognises the literal_str class).
def literal_str_representer(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
# Add to yaml.
yaml.add_representer(literal_str, literal_str_representer)

class flow_mapping(dict): pass
def flow_mapping_representer(dumper, data):
    return dumper.represent_mapping( u'tag:yaml.org,2002:map', data, flow_style=True )

yaml.add_representer(flow_mapping, flow_mapping_representer)
    
###########
# Classes #
###########

class Digichem_coords_ABC(Input_file, Molecule_mixin):
    """
    ABC for classes that represents an input file in the digichem input (.si) format.
    """

    @classmethod
    def from_xyz(self, geometry, **kwargs):
        """
        Create a Digichem_coords object from a molecule in xyz format.
        
        :param geometry: The input geometry in xyz format.
        :param charge: The molecular charge.
        :param multiplicity: The molecular multiplicity (as an integer).
        :param name: Name of the system/molecule
        """
        raise NotImplementedError("Implement in subclass")
    
    @classmethod
    def from_result(self, result, **kwargs):
        if not kwargs.get('charge'):
            kwargs['charge'] = result.atoms.charge
        
        if not kwargs.get('multiplicity'):
            kwargs['multiplicity'] = result.metadata.multiplicity
            
        if not kwargs.get('history'):
            # Note it's not the history of the old calc we want here, this old calc IS our new history.
            kwargs['history'] = result._id
            
        if not kwargs.get('file_name') and len(result.metadata.log_files):
            kwargs['file_name'] = result.metadata.log_files[0]
        
        if not kwargs.get('name'):
            kwargs['name'] = result.metadata.name
        
        return self.from_xyz(
            result.atoms.to_xyz(),
            **kwargs
        )
    
    def dump(self):
        """
        Get this input file as a dict.
        """
        dic = {
            'version': "2.1.0",
        }
        dic.update(super().dump())
        dic['atoms'] = self.atoms
        return dic
        
    @property
    def dict(self):
        """
        Get this input file as a dict.
        """
        return self.dump()
        
    @property
    def element_dict(self):
        """
        Get a dictionary where each key is one of the elements in this molecules (C, H, N etc) and the value is the number of that element that appears in the molecule
        
        :return: The element dictionary.
        """
        elements = {}
        for atom in self.atoms:
            # Try and increment the count of the atom.
            try:
                elements[atom['atom']] += 1
            except KeyError:
                # Add the new atom.
                elements[atom['atom']] = 1
        return elements
        
    @property
    def yaml(self):
        """
        Get this input file in yaml format.
        """
        # Get in dict format.
        intermediate = self.dict
        
        # Wrap geometry in literal format (so it will appear line by line).
        intermediate['atoms'] = [flow_mapping(atom) for atom in intermediate['atoms']]
        
        # Convert.
        return yaml.dump(intermediate, sort_keys=False)
        
    @property
    def xyz(self):
        """
        Get the geometry of this input file in XYZ format.
        """
        return "{}\n\n{}".format(len(self.geometry.strip().split("\n")), self.geometry)
    
    @property
    def elements(self):
        """
        A unique list of the elements in this input file.
        
        The elements are returned as a list of integers (atomic numbers)
        """
        return list(set(
            periodictable.elements.symbol(coord['atom']).number for coord in self.atoms
        ))
    
    @classmethod
    def from_com(self, geometry, *, charge = None, multiplicity = None, **kwargs):
        """
        Create a Digichem_coords object from a molecule in gaussian input format (.com, .gjf etc).
        
        :param geometry: The input geometry in gaussian format.
        :param charge: The molecular charge.
        :param multiplicity: The molecular multiplicity (as an integer).
        :param name: Name of the system/molecule
        """
        # Get a parser for our gaussian input file.
        parser = Gaussian_input_parser(geometry)
        
        # If we've not been given a charge and/or multi, use the values from the input file.
        if charge is None:
            charge = parser.charge
        if multiplicity is None:
            multiplicity = parser.multiplicity
            
        # Continue with other constructors.
        return self.from_xyz(parser.xyz, charge = charge, multiplicity = multiplicity, **kwargs)
    
    @classmethod
    def from_yaml(self, yaml_dict, file_name = None, **kwargs):
        """
        Create a Digichem_coords object from a loaded/parsed .si file.
        """
        yaml_dict = dict(yaml_dict)
        # Overwrite dictionary values if explicit ones have been given.
        for kwarg in kwargs:
            if kwargs[kwarg] is not None:
                yaml_dict[kwarg] = kwargs[kwarg]
            
        # Continue constructing.
        try:
            return self(**yaml_dict, file_name = file_name)
        except TypeError:
            raise Digichem_exception("Failed to load .si file '{}'; is the file formatted correctly?".format(file_name))
    
    @classmethod
    def from_si(self, si_file, **kwargs):
        """
        Create a Digichem_coords object from a raw .si file.
        """
        return self.from_yaml(yaml.safe_load(si_file), **kwargs)
        
    def to_file(self, file):
        """
        Write this digichem input file to an open file object.
        
        :param file: A file opened with open() (or similar).
        """
        file.write(self.yaml)
        
    def to_format(self, file_type, file = None):
        """
        Get this input file in an arbitrary format.
        
        :param file_type: The format of the file; see output_formats()
        :param file: An optional file to write to, if not given the converted file is returned as a string.
        """
        if file_type.lower() == "si":
            # Convert to yaml
            return self.yaml
        else:
            # Convert.
            charge = self.charge if self.charge is not None else None
            multiplicity = self.multiplicity if self.multiplicity is not None else None
            return Openprattle_converter(input_file_buffer = self.xyz, input_file_path = self.implicit_name, input_file_type = "xyz").convert(file_type, file, charge = charge, multiplicity = multiplicity)

    @classmethod
    def input_formats(self):
        """
        A dictionary of available input formats that this file can be created from.
        
        Each key is the short-code of the format (eg, si, com, xyz etc) while the value is a longer description.
        """
        formats = {
            "si": "Digichem Input Format",
            "com": "Gaussian Input",
            "gau": "Gaussian Input",
            "gjc": "Gaussian Input",
            "gjf": "Gaussian Input"
        }
        formats.update(Oprattle_formats().read())
        
        # Some formats supported by obabel don't make sense for us to use.
        formats.pop('text', None)
        formats.pop('txt', None)
        return formats

    @classmethod
    def output_formats(self):
        """
        A dictionary of available output formats that this file can be converted to.
        
        Each key is the short-code of the format (eg, si, com, xyz etc) while the value is a longer description.
        """
        formats = {"si": "Digichem Input Format"}
        formats.update(Oprattle_formats().write())
        
        # Some formats supported by obabel don't make sense for us to use.
        formats.pop('copy', None)
        formats.pop('k', None)
        formats.pop('confabreport', None)
        formats.pop('nul', None)
        formats.pop('text', None)
        formats.pop('txt', None)
        
        return formats
    
    def __eq__(self, other):
        """
        Check for equality with another coord object.
        """
        try:
            assert self.charge == other.charge
            assert self.multiplicity == other.multiplicity
            
            # TODO: This check should not care about the order of atoms.
            for index, atom in enumerate(self.atoms):
                assert atom == other.atoms[index]
            
            return True
        
        except AssertionError:
            return False


class Digichem_coords_v2(Digichem_coords_ABC):
    """
    Class that represents an input file in the digichem input (.si) format (V2).
    
    The .si format is YAML based and stores atom positions in a dictionary format along with charge and multiplicity.
    """
    
    def __init__(self, atoms, *, charge = None, multiplicity = None, name = None, version = "2.1.0", file_name = None, history = None):
        """
        Constructor for .si files.
         
        :param atoms: A list of dictionaries specifying the geometry. Each dict should contain four keys, 'atom': The element, 'x': The x coord, 'y': The y coord, 'z': The z coord. All coordinates are in angstroms.
        :param charge: The molecular charge (as an integer).
        :param multiplicity: The molecular multiplicity (as an integer).
        :param name: Name of the system/molecule.
        :param version: The version of the .si file.
        :param file_name: The name of the file which was loaded. This can be used as a back-up file name.
        """
        super().__init__(charge, multiplicity, name, file_name, history = history)
        self.atoms = atoms
        self.version = version
        
    @classmethod
    def from_xyz(self, geometry, **kwargs):
        """
        Create a Digichem_coords object from a molecule in xyz format.
        
        :param geometry: The input geometry in xyz format.
        :param charge: The molecular charge.
        :param multiplicity: The molecular multiplicity (as an integer).
        :param name: Name of the system/molecule
        """
        # Split the xyz format on newlines.
        split_geom = geometry.strip().split("\n")
        
        # Remove the first line (the number of atoms) and the second (optional comment line)
        split_geom = split_geom[2:]
        
        atoms = []
        for geom_line in split_geom:
            atom, x_coord, y_coord, z_coord = geom_line.split()
            atoms.append({"atom": atom, "x": float(x_coord), "y": float(y_coord), "z": float(z_coord)})
        
        # Call our main constructor.
        return self(atoms, **kwargs)
        
    @property
    def geometry(self):
        """
        Get the geometry of this input file in XYZ format.
        """
        # Build the geometry section.
        # XYZ is a little bit loosely defined.
        # Obabel (defacto standard? uses ~10 chars for each column?).
        # That's 9 usable chars +1 obligate space.
        geometry = []
        for atom in self.atoms:
            geometry.append("{:<9} {:>14.8f} {:>14.8f} {:>14.8f}".format(atom['atom'], atom['x'], atom['y'], atom['z']).strip())
            
        return "\n".join(geometry)


class Digichem_coords_v1(Digichem_coords_ABC):
    """
    Class that represents an input file in the digichem input (.si) format (V1).
     
    The .si format is yaml based, and stores the input geometry in a modified xyz format along with charge and multiplicity.
    """
    
    def __init__(self, geometry, *, charge = None, multiplicity = None, name = None, version = "1.0.0", file_name = None):
        """
        Constructor for .si files.
         
        :param geometry: The molecular geometry in .si format. Use from_xyz() instead if your format is in xyz.
        :param charge: The molecular charge (as an integer).
        :param multiplicity: The molecular multiplicity (as an integer).
        :param name: Name of the system/molecule.
        :param file_name: The name of the file which was loaded. This can be used as a back-up file name.
        """
        super().__init__(charge, multiplicity, name, file_name)
        self.geometry = geometry
        self.version = version

    @classmethod
    def from_xyz(self, geometry, **kwargs):
        """
        Create a Digichem_coords object from a molecule in xyz format.
        
        :param geometry: The input geometry in xyz format.
        :param charge: The molecular charge.
        :param multiplicity: The molecular multiplicity (as an integer).
        :param name: Name of the system/molecule
        """
        # Split the xyz format on newlines.
        split_geom = geometry.split("\n")
        
        # Remove the first line (the number of atoms) and the second (optional comment line)
        split_geom = split_geom[2:]
        
        # Call our main constructor.
        return self("\n".join(split_geom), **kwargs)
    
    @property
    def atoms(self):
        return Digichem_coords_v2.from_xyz(self.xyz, charge = self.charge, multiplicity = self.multiplicity, name = self.name, file_name = self.file_name).atoms


####################
# Helper utilities #
####################

Digichem_coords = Digichem_coords_v2

def si_from_yaml(yaml_dict, file_name = None, **kwargs):
    """
    """
    # In the first instance, look to see if there's a version string.
    if "version" in yaml_dict:
        version = packaging.version.parse(yaml_dict['version'])
        
        if version < packaging.version.parse("2"):
            # V1.
            cls = Digichem_coords_v1
        
        elif version < packaging.version.parse("3"):
            # V2.
            cls = Digichem_coords_v2
        
        else:
            raise Digichem_exception("Unsupported .si file version '{}'".format(yaml_dict['version']))
    
    else:
        # No explicit version.
        # If we have the old 'geometry' section, assume v1.
        if "geometry" in yaml_dict:
            cls = Digichem_coords_v1
            
        else:
            # Assume the latest format.
            cls = Digichem_coords    
            
    return cls.from_yaml(yaml_dict, file_name, **kwargs)


def si_from_data(data, file_type, *, gen3D = None, file_name = None, **kwargs):
    """
    """
    if file_type in ["com", "gau", "gjc", "gjf"]:
        # Gaussian input format.
        return Digichem_coords.from_com(data, file_name = file_name, **kwargs)

    elif file_type == "si":
        # Digichem input format.
        return si_from_yaml(yaml.safe_load(data), file_name = file_name, **kwargs)
        
    elif file_type == "pickle":
        # A digichem resume file.
        # The resume file (should be) a pickled destination object.
        try:
            destination = dill.loads(data)
        
        except Exception as e:
            raise Digichem_exception("Failed to parse digichem resume file") from e
        
        return destination.program.calculation.input_coords
    
    else:
        # Generic input format, use obabel.
        
        # We convert all formats to gaussian input formats (because this format contains charge and multiplicity, which we can extract).
        com_file = Openprattle_converter(input_file_buffer = data, input_file_type = file_type).convert("com", gen3D = gen3D)
    
        # Continue with other constructors.
        return Digichem_coords.from_com(com_file, file_name = file_name, **kwargs)


def si_from_file(file_name, file_type = None, *, gen3D = None, **kwargs):
        """
        Create a Digichem_coords object from a file in arbitrary format.
        
        :param file_name: Name/path of the input file to read from.
        :param file_type: The format of the file; a string recognised by Digichem_coords.input_formats(). If not given, an attempt will be made to guess from the file name.
        :param charge: The molecular charge.
        :param multiplicity: The molecular multiplicity (as an integer).
        :param name: Name of the system/molecule.
        """
        file_name = Path(file_name)
        digichem.log.get_logger().info("Parsing coordinate file '{}'".format(file_name))
        auto_file_type = False
        
        try:
            # Get the file format.
            if file_type is None:
                auto_file_type = True
                file_type = Openprattle_converter.type_from_file_name(file_name, allow_none = True)
                    
            # Certain formats we support natively; others we convert to an intermediate format.                
            # NOTE: Here we assume files without an extension are log files.
            # This works fine for directories, but might change in future.
            if file_type in ["dat", "log", "out", "output", None] \
                or (auto_file_type and "".join(file_name.suffixes) in open_for_parsing.get_archive_formats()):
                # Generic log-file (output) format.
                # Most formats (.log, .dat etc) we can parse with either Obabel or Digichem.
                # Some broken log files (incomplete) we can parse only with Obabel.
                # Some unusual formats (directories, archives) we can parse only with Digichem.
                
                # Try with Digichem first.
                try:
                    result = parse_calculation(file_name, options = digichem.config.get_config(), format_hint = "cclib")
                    return Digichem_coords.from_result(result, file_name = file_name, **kwargs)
                
                except Exception as e:
                    # No good, see if we can use obabel.
                    try:
                        com_file = Openprattle_converter(file_name, file_type).convert("com", gen3D = gen3D)
                    
                    except Exception:
                        # Also no good, re-raise original exception.
                        raise e
                    
                    # Worked with fallback, log a message.
                    digichem.log.get_logger().warning(f"Failed to parse calculation output file '{file_name}'; using Obabel fallback mechanism")
                    return Digichem_coords.from_com(com_file, file_name = file_name, **kwargs)
                
            else:
                with open(file_name, "rb" if file_type in ["pickle", "cdx"] else "r") as input_file:
                    return si_from_data(input_file.read(), file_name = file_name, gen3D = gen3D, file_type = file_type, **kwargs)
            
        except:
            raise ValueError("Could not parse coordinates from '{}'".format(file_name))
    


import yaml
from deepmerge import Merger
from pathlib import Path

from configurables import Configurable, Options, Option
from configurables.option import Nested_dict_type

from digichem.config.locations import user_config_location
from digichem.misc.io import atomic_write
from digichem.translate import Cube_grid_points


class Auto_type():
    """
    A class that when called, converts a string into a more appropriate type automatically.
    
    The rules for deciding which type to convert to are the same as for parsing yaml using pyyaml, as that is the module that is relied on for conversion.
    """
    
    def __new__(cls, value):
        return yaml.safe_load(value)


class Digichem_options(Configurable):
    """
    Class for holding main digichem options from various sources.
    """
    
    alignment = Option(help = "The default alignment method to use, MIN: minimal, FAP: furthest atom pair, AA: average angle, AAA: advanced average angle.", choices = ["MIN", "FAP", "AA", "AAA"], default = "MIN")
    angle_units = Option(help = "The default angle units to use, deg: degrees, rad: radians.", choices = ["deg", "rad"], default = "deg")
    
    external = Options(
        help = "Options specifying paths to various external programs that digichem may use. If no path is given, then these programs will simply be executed by name (so relying on OS path resolution to find the necessary executables, which is normally fine.)",
        formchk = Option(help = "Gaussian's formchk utility https://gaussian.com/formchk/", default = "formchk"),
        cubegen = Option(help = "Gaussian's cubegen utility https://gaussian.com/cubegen/", default = "cubegen"),
        cubegen_parallel = Option(help = "What type of parallelism to use with cubegen, multithreaded runs a single instance of cubegen across multiple CPUs, pool runs multiple instances of cubegen", choices = [None, "multithreaded", "pool"], default = "pool")
    )

    parse = Options(
        help = "Options for controlling parsing of config files",
        profiling_rows = Option(help = "The maximum number of rows to parse from the calculation profile file (if available); if more rows than this are available then the data will be downsampled to at most this number of data points", type = int, default = 1000)
    )
    
    skeletal_image = Options(
        help = "Options for controlling the rendering of 2D skeletal images.",
        resolution = Options(help = "The resolution for rendered images", 
            absolute = Option(help = "The output resolution (x and y) in pixels", type = int, default = None),
            relative = Option(help = "The output resolution (x and y) in multiples of the length of the molecule", type = int, default = 100)
        )
    )
    
    logging = Options(
        help = "Options relating to output of error messages. Note that the final logging level is determined by combining both 'log_level' and 'verbose', so a 'log_level' of 'OFF' and 'verbose' of '2' is equal to 'ERROR'.",
        log_level = Option(help = "The level of messages to output, one of OFF (no logging at all), CRITICAL (fewest messages), ERROR, WARNING, INFO or DEBUG (most messages)", choices = ["OFF", "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"], default = "INFO"),
        verbose = Option(help = "Increase the verbosity of the program by this amount. Each integer increase of verbosity will increase 'log_level' by 1 degree.", type = int, default = 0),
        render_logging = Option(help = "Whether to print output from render engines.", type = bool, default = False)
    )
    
    render = Options(
        help = "Options for controlling the appearance of 3D molecule images.",
        enable_rendering = Option(help = "Set to False to disable image rendering.", type = bool, default = True),
        engine = Option(help = "The rendering engine to use", choices = ["vmd", "batoms"], default = "vmd"),
        vmd = Options(help = "VMD specific options (only applies if engine == 'vmd'",
            executable = Option(help = "Path to the VMD (Visual Molecular Dynamics) executable", default = "vmd"),
            tachyon = Option(help = "The tachyon ray-tracing library, performs the actual rendering. Tachyon is typically packaged with VMD, but often isn't added to the path automatically", default = "tachyon"),
            rendering_style = Option(help =\
"""The render/display mode, changes the appearance of rendered molecules/orbitals.
Possible options are:
 pastel: The default style, uses light, pastel-ish colours for orbitals with low transparency. Normal atom colours.
 light-pastel: Similar to pastel, but with lighter orbital colours.
 dark-pastel: Similar to pastel, but with darker orbital colours, reminiscant of the 'sharp' style.
 sharp: Orbitals are darkly coloured around their edge but have high transparency in their center. Normal atom colours.
 gaussian: A style attempting to mimic that of GaussView.
 vesta: A style attempting to mimic that of VESTA.""",
                choices = ["pastel", "light-pastel", "dark-pastel", "sharp", "gaussian", "vesta"], default = "pastel"
            ),
        ),
        batoms = Options(help = "Beautiful Atoms/Blender specific options (only applies if engine == 'batoms'",
            blender = Option(help = "Path to the blender executable, in which beautiful atoms should be installed", default = "batoms-blender"),
            cpus = Option(help = "The number of CPUs/threads to use. This option is overridden if running in a calculation environment (where it uses the same number of CPUs as the calculation did)", type = int, default = 1),
            render_samples = Option(help = "The number of render samples (or passes) to use. Higher values result in higher image quality and greater render times", type = int, default = 32),
            perspective = Option(help = "The perspective mode", choices = ["orthographic", "perspective"], default = "perspective"),
            stacking = Option(help = "The number of image copies to composite together to avoid transparency artifacts", type = int, default = 10)
            # TODO: Colour options.
        ),
        safe_cubes = Option(help = "Whether to sanitize cubes so older software can parse them (VMD < 1.9.2 etc)", type = bool, default = False),
        use_existing = Option(help =\
"""If True, previously created files will be reused. If False, new images will rendered, replacing the old.
This is on by default for 3D rendered images because they are expensive (time-consuming) to render.""", type = bool, default = True
        ),
        
        auto_crop = Option(help = "Whether to enable  automatic cropping of excess whitespace around the border of generated images.  If False, overall image rendering is likely to take less time, but molecules may only occupy a small portion of the true image.", type = bool, default = True),
        resolution = Option(help = "The target resolution for rendered images. Higher values will increase image quality, at the cost of increased render time and file size.", type = int, default = 512),
        orbital = Options(help = "Specific options for orbital density plots.",
            cube_grid_size = Option(help =\
"""The size of grid used to generate cube files.
Densities are plotted on a 3D grid of points, this option controls how many points there are in the grid (per dimension).
In addition to an integer number of points (~100 is often sufficient), any of the following keywords can also be specified:
Tiny, Small, Medium, Large, Huge or Default.""", type = Cube_grid_points, default = Cube_grid_points("Default")
            ),
            isovalue = Option(help = "The isovalue to use for rendering orbital density.", type = float, default = 0.02)
        ),
        spin = Options(help = "Specific options for spin density plots.",
            cube_grid_size = Option(help = "The size of the grid use to plot cube data. As cubes of this type are rendered with a smaller isovalue, it is often necessary to use a larger grid size than normal to maintain quality.", type = Cube_grid_points, default = Cube_grid_points("Large")),
            isovalue = Option(help = "The isovalue to use for plotting spin density.", type = float, default = 0.0004)
        ),
        density = Options(help = "Specific options for total density plots.",
            cube_grid_size = Option(help = "The size of the grid use to plot cube data.", type = Cube_grid_points, default = Cube_grid_points("Large")),
            isovalue = Option(help = "The isovalue to use for plotting total density.", type = float, default = 0.02)
        ),
        difference_density = Options(help = "Specific options for excited states difference density plots.",
            cube_grid_size = Option(help = "The size of the grid use to plot cube data.", type = Cube_grid_points, default = Cube_grid_points("Default")),
            isovalue = Option(help = "The isovalue to use for difference density plots.", type = float, default = 0.001)
        ),
        natural_transition_orbital = Options(help = "Specific options for NTO plots.",
            cube_grid_size = Option(help = "The size of the grid use to plot cube data.", type = Cube_grid_points, default = Cube_grid_points("Default")),
            isovalue = Option(help = "The isovalue to use for NTO plots.", type = float, default = 0.02)
            
        ),
        dipole_moment = Options(help = "Specific options for permanent dipole moment plots.",
            scaling = Option(help = "The value (x) to scale the TDM by, where 1 D = x Å.", type = float, default = 1.0)
            
        ),
        transition_dipole_moment = Options(help = "Specific options for transition dipole moment plots.",
            electric_scaling = Option(help = "The value (x) to scale the TEDM by, where 1 D = x Å.", type = float, default = 5.0),
            magnetic_scaling = Option(help = "The value (x) to scale the TMDM by, where 1 au = x Å.", type = float, default = 10.0),
        ),
            
    )
    
    orbital_diagram = Options(help = "Options that control the appearance of orbital energy diagrams.",
        enable_rendering = Option(help = "Set to False to disable image rendering.", type = bool, default = True),
        y_limits = Option(help =\
"""The method used to set the min/max limits of the Y axis.
Possible options are:
  - 'all': Limits are set so all orbitals are visible.
  - 'auto': Limits are set automatically so the 0 point (ie, 'X' axis), HOMO and LUMO are clearly visible.
  - 'center': Like 'auto' except the lower limit is set more negative so the HOMO-LUMO are closer to the middle of the diagram.
  - a list of [y_min, y_max], where y_min is the most negative value on the Y axis, and y_max is the most positive value on the Y axis (both in eV).""", type = Auto_type, default = "auto" 
        ),
        full_axis_lines = Option(help = "If True, black lines are drawn around the boarder of the diagram. If False, a line is drawn for the Y axis but not for the other 3 sides of the diagram.", type = bool, default = False)
    )
    
    excited_states_diagram = Options(help = "Options that control the appearance of excited states energy diagrams.",
        enable_rendering = Option(help = "Set to False to disable image rendering.", type = bool, default = True),
        y_limits = Option(help =\
"""The method used to set the min/max limits of the Y axis.
Possible options are:
  - 'all': Limits are set so all excited states are visible.
  - 'auto': Limits are set automatically so the lowest excited state of each multiplicity is clearly visible (S1, D1, T1, Q1... N1 etc).
  - a list of [y_min, y_max], where y_min is the most negative value on the Y axis, and y_max is the most positive value on the Y axis (both in eV).""", type = Auto_type, default = "all"
        ),
        show_dest = Option(help = "Whether or not to show the dE(ST) label.", type = bool, default = True)
    )
    
    absorption_spectrum = Options(help = "Options for controlling the appearance of simulated UV-Vis like absorption spectra.",
        enable_rendering = Option(help = "Set to False to disable image rendering.", type = bool, default = True),
        y_limits = Option(help =\
"""The method used to set the min/max limits of the Y axis.
Possible options are:
  - 'auto': Limits are set automatically so all peaks are clearly visible.
  - a list of [y_min, y_max], where y_min is the most negative value on the Y axis, and y_max is the most positive value on the Y axis (both in oscillator strength).""", type = Auto_type, default = "auto"
        ),
        x_limits = Option(help =\
"""The method used to set the min/max limits of the X axis.
Possible options are:
  - 'auto': Limits are set automatically so all peaks are clearly visible.
  - a list of [x_min, x_max], where x_min is the most negative value on the X axis, and x_max is the most positive value on the X axis (both in nm).""", type = Auto_type, default = "auto"
        ),
        max_width = Option(help =\
"""The maximum image width in pixels.
Absorption graphs will grow/shrink their width to fit available data, keeping a constant scale (constant pixels to nm ratio) but only up to this maximum.
To disable the maximum width, set to null.""", type = int, default = 1200
        ),
        peak_cutoff = Option(help =\
"""The minimum oscillator strength that a peak must have to be shown in the graph, as a fraction ofthe highest peak.
Set to 0 for no cutoff (all peaks shown), which may results in the graph being extended well beyond the drawn peaks (because many peaks are too small to see).
This option has no effect when using manual x limits.""", type = float, default = 0.01
        ),
        y_filter = Option(help = "The minimum y value to simulate using the Gaussian function (y values below this are discarded)", type = float, default = 1e-6),
        x_padding = Option(help = "The amount (in nm) to extend the x axis past the highest/lowest energy peak.", type = int, default = 40),
        fwhm = Option(help = "The full-width at half-maximum; changes how wide the drawn peaks are. Note that the choice of peak width is essentially arbitrary; only the peak height is given by calculation. Units are eV.", type = float, default = 0.4),
        gaussian_cutoff = Option(help = "The minimum y value to plot using the Gaussian function (controls how close to the x axis we draw the gaussian) as a fraction of the max peak height.", type = float, default = 0.001),
        gaussian_resolution = Option(help = "The spacing between x values to plot using the Gaussian function, in eV. Values that are too large will result in 'curves' made up of a series of straight edges.", type = float, default = 0.01),
        use_jacobian = Option(help = "Whether or not to use the jacobian transformation to correctly scale the y-axis (see J. Phys. Chem. Lett. 2014, 5, 20, 3497)", type = bool, default = True),
        plot_bars = Option(help = "Whether to plot vertical bars for each excited state.", type = bool, default = True),
        plot_peaks = Option(help = "Whether to plot individual Gaussian functions for each excited state.", type = bool, default = False),
        plot_cumulative_peak = Option(help = "Whether to plot the sum of all Gaussian functions (most closely simulates a real spectrum).", type = bool, default = True)
    )
    
    emission_spectrum = Options(help = "Options for controlling the appearance of simulated emission spectra. 'emission_spectrum' and 'absorption_spectrum 'take the same options.",
        enable_rendering = Option(help = "Set to False to disable image rendering.", type = bool, default = True),
        y_limits = Option(help =\
"""The method used to set the min/max limits of the Y axis.
Possible options are:
  - 'auto': Limits are set automatically so all peaks are clearly visible.
  - a list of [y_min, y_max], where y_min is the most negative value on the Y axis, and y_max is the most positive value on the Y axis (both in oscillator strength).""", type = Auto_type, default = "auto"
        ),
        x_limits = Option(help =\
"""The method used to set the min/max limits of the X axis.
Possible options are:
  - 'auto': Limits are set automatically so all peaks are clearly visible.
  - a list of [x_min, x_max], where x_min is the most negative value on the X axis, and x_max is the most positive value on the X axis (both in nm).""", type = Auto_type, default = "auto"
        ),
        max_width = Option(help =\
"""The maximum image width in pixels.
Absorption graphs will grow/shrink their width to fit available data, keeping a constant scale (constant pixels to nm ratio) but only up to this maximum.
To disable the maximum width, set to null.""", type = int, default = 1200
        ),
        peak_cutoff = Option(help =\
"""The minimum oscillator strength that a peak must have to be shown in the graph, as a fraction of the highest peak.
Set to 0 for no cutoff (all peaks shown), which may results in the graph being extended well beyond the drawn peaks (because many peaks are too small to see).
This option has no effect when using manual x limits.""", type = float, default = 0.01
        ),
        y_filter = Option(help = "The minimum y value to simulate using the Gaussian function (y values below this are discarded)", type = float, default = 1e-6),
        x_padding = Option(help = "The amount (in nm) to extend the x axis past the highest/lowest energy peak.", type = int, default = 40),
        fwhm = Option(help = "The full-width at half-maximum; changes how wide the drawn peaks are. Note that the choice of peak width is essentially arbitrary; only the peak height is given by calculation. Units are eV.", type = float, default = 0.4),
        gaussian_cutoff = Option(help = "The minimum y value to plot using the Gaussian function (controls how close to the x axis we draw the gaussian) as a fraction of the max peak height.", type = float, default = 0.001),
        gaussian_resolution = Option(help = "The spacing between x values to plot using the Gaussian function, in eV. Values that are too large will result in 'curves' made up of a series of straight edges.", type = float, default = 0.01),
        use_jacobian = Option(help = "Whether or not to use the jacobian transformation to correctly scale the y-axis (see J. Phys. Chem. Lett. 2014, 5, 20, 3497)", type = bool, default = True),
        plot_bars = Option(help = "Whether to plot vertical bars for each excited state.", type = bool, default = True),
        plot_peaks = Option(help = "Whether to plot individual Gaussian functions for each excited state.", type = bool, default = False),
        plot_cumulative_peak = Option(help = "Whether to plot the sum of all Gaussian functions (most closely simulates a real spectrum).", type = bool, default = True)
    )
    
    IR_spectrum = Options(help = "Options for controlling the appearance of simulated IR like vibrational frequency spectra.",
        enable_rendering = Option(help = "Set to False to disable image rendering.", type = bool, default = True),
        y_limits = Option(help =\
"""The method used to set the min/max limits of the Y axis.
Possible options are:
  - 'auto': Limits are set automatically so all peaks are clearly visible.
  - a list of [y_min, y_max], where y_min is the most negative value on the Y axis, and y_max is the most positive value on the Y axis (both in km/mol).""", type = Auto_type, default = "auto"
        ),
        x_limits = Option(help =\
"""The method used to set the min/max limits of the X axis.
Possible options are:
  - 'auto': Limits are set automatically so all peaks are clearly visible.
  - a list of [x_min, x_max], where x_min is the most negative value on the X axis, and x_max is the most positive value on the X axis (both in cm-1).""", type = Auto_type, default = "auto"
        ),
        fwhm = Option(help = "The full-width at half-maximum; changes how wide the drawn peaks are. Note that the choice of peak width is essentially arbitrary; only the peak height is given by calculation. Units are cm-1.", type = float, default = 80),
        max_width = Option(help =\
"""The maximum image width in pixels.
IR spectra will grow/shrink their width to fit available data, keeping a constant scale (constant pixels to nm ratio) but only up to this maximum.
To disable the maximum width, set to null.""", type = int, default = 1500),
        y_filter = Option(help = "The minimum y value to simulate using the Gaussian function (y values below this are discarded)", type = float, default = 1e-6),
        gaussian_cutoff = Option(help = "The minimum y value to plot using the Gaussian function (controls how close to the x axis we draw the gaussian) as a fraction of the max peak height.", type = float, default = 0.001),
        gaussian_resolution = Option(help = "The spacing between x values to plot using the Gaussian function, in eV. Values that are too large will result in 'curves' made up of a series of straight edges.", type = float, default = 1.0)
    )
    
    nmr = Options(help = "Options for controlling simulated NMR spectra",
        enable_rendering = Option(help = "Set to False to disable image rendering.", type = bool, default = True),
        coupling_filter = Option(help = "Discard J coupling that is below this threshold (in Hz)", type = float, default = 1),
        fwhm = Option(help = "The full-width at half-maximum; changes how wide the drawn peaks are. Note that the choice of peak width is essentially arbitrary; only the peak height is given by calculation. Units are ppm.", type = float, default = 0.01),
        y_filter = Option(help = "The minimum y value to simulate using the Gaussian function (y values below this are discarded)", type = float, default = 1e-6),
        gaussian_cutoff = Option(help = "The minimum y value to plot using the Gaussian function (controls how close to the x axis we draw the gaussian) as a fraction of the max peak height.", type = float, default = 0.001),
        gaussian_resolution = Option(help = "The spacing between x values to plot using the Gaussian function, in ppm. Values that are too large will result in 'curves' made up of a series of straight edges.", type = float, default = 0.001),
        frequency = Option(help = "The frequency to run the simulated spectrometer at. Larger values will result in narrower coupling. Units are MHz", type = float, default = 100),
        pre_merge = Option(help = "The threshold within which similar peaks will be merged together, before they are split by coupling. This option typically results in faster execution, but more error, compared to post_merge. Units in ppm.", type = float, default = 0.01),
        post_merge = Option(help = "The threshold within which similar peaks will be merged together. Units in ppm.", type = float, default = None),
        standards = Option(help = "The chemical shift of a standard reference peak (in ppm) to use to adjust the spectrum. One for each element.", type = Nested_dict_type, default = Nested_dict_type({
            "H": 31.68766666666667,
            "B": 111.27199999999999,
            "C": 197.90316666666664,
            "F": 183.2779999999999,
            "P": 290.8630000000001,
        })),
        plot_bars = Option(help = "Whether to plot vertical bars for each NMR shift", type = bool, default = False),
        plot_zoom_bars = Option(help = "Whether to plot vertical bars for each NMR shift in zoomed spectra", type = bool, default = True),
        
        # TODO: Validation on these sorts of options is poor and needs looking at.
        isotopes = Option(help = "Isotope specific options. Each key should consist of a tuple of (proton_number, isotope).", type = Nested_dict_type, default = Nested_dict_type({
            # Resonance frequencies calculated at 9.3947 T.
            # 1H, increase fidelity to see more detail.
            "1H": {"frequency": 400, "fwhm": 0.005, "gaussian_resolution": 0.0005, "coupling_filter": 0.001, "pre_merge": 0.0005},
            # 11B.
            "11B": {"frequency": 128.3},
            # 13C.
            "13C": {"frequency": 100.6},
            # 19F
            "19F": {"frequency": 376.5},
            # 31P
            "31P": {"frequency": 162.0}
        }))
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, allow_unrecognised_options = True, **kwargs)
            
    @classmethod
    def _from_reduce(self, kwargs):
        """
        Function to help un-pickling this class.
        """
        return self(**kwargs)
    
    def __reduce__(self):
        """
        Magic function to help pickling this class.
        """
        return (self._from_reduce, (self.dump(),))
        
    # These methods allow the main digichem options object to be accessed as a dict.
    # This is largely to provide compatibility with legacy code, where the digichem options object was, in fact, a dict.
    def __getitem__(self, key):
        return getattr(self, key)
    
    def __setitem__(self, key, value):
        setattr(self, key, value)
        
    def __delitem__(self, key):
        # Deleting a configurable option doesn't actually delete it (because that wouldn't make much sense),
        # Instead it reverts that option to its default.
        delattr(self, key)
    
    def save(self, path = user_config_location):
        """
        Save the current value of these options to file, so that they will be reloaded on next program start.
        
        :param path: Where to save to (the user's config file by default).
        """
        data = yaml.dump(self.dump())

        path = Path(path)
        
        try:
            path.parent.mkdir(exist_ok = True)
            atomic_write(path, data)
        
        except FileNotFoundError as e:
            # We lost the race, give up.
            raise Exception("Failed to write settings to file '{}'; one of the parent directories does not exist".format(path)) from e
    



import pytest
from pathlib import Path

from digichem.image.excited_states import Excited_states_diagram_maker
from digichem.image.graph import Convergence_graph_maker
from digichem.image.orbitals import Orbital_diagram_maker
from digichem.image.spectroscopy import Absorption_graph_maker, Emission_graph_maker, \
    Frequency_graph_maker, NMR_graph_maker, NMR_graph_zoom_maker
from digichem.image.structure import Skeletal_image_maker
from digichem.image.vmd import Structure_image_maker, Orbital_image_maker, Density_image_maker, \
    Combined_orbital_image_maker, Permanent_dipole_image_maker, Transition_dipole_image_maker
from digichem.file.cube import Fchk_to_cube, Fchk_to_density_cube, Fchk_to_nto_cube, Fchk_to_spin_cube
from digichem.parse.util import open_for_parsing, parse_calculation

from digichem.test.util import digichem_options, data_directory
from digichem.test.test_result import gaussian_ES_result, turbomole_ES_result, orca_ES_result, \
    gaussian_opt_result, turbomole_opt_result, orca_opt_result, orca_opt_freq_result, \
    orca_nmr_result

pytest.skip(allow_module_level=True)

@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_ES_result"),
    pytest.lazy_fixture("turbomole_ES_result"),
    pytest.lazy_fixture("orca_ES_result")
])
def test_es_diagram(result_set, tmp_path, digichem_options):
    """Can we make an excited states diagram?"""
    maker = Excited_states_diagram_maker.from_options(
        tmp_path / "tmp.png",
        excited_states = result_set.excited_states,
        ground_state = result_set.ground_state,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_opt_result"),
    pytest.lazy_fixture("turbomole_opt_result"),
    pytest.lazy_fixture("orca_opt_result")
])
def test_convergence_diagram(result_set, tmp_path, digichem_options):
    """Can we make an optimisation graph?"""
    maker = Convergence_graph_maker.from_options(
        tmp_path / "tmp.png",
        energies = result_set.energies.scf,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_opt_result"),
    pytest.lazy_fixture("turbomole_opt_result"),
    pytest.lazy_fixture("orca_opt_result")
])
def test_orbital_diagram(result_set, tmp_path, digichem_options):
    """Can we make an orbital diagram?"""
    maker = Orbital_diagram_maker.from_options(
        tmp_path / "tmp.png",
        orbitals = result_set.orbitals,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_ES_result"),
    pytest.lazy_fixture("turbomole_ES_result"),
    pytest.lazy_fixture("orca_ES_result")
])
def test_abs_diagram(result_set, tmp_path, digichem_options):
    """Can we make an absorption spectrum?"""
    maker = Absorption_graph_maker.from_options(
        tmp_path / "tmp.png",
        excited_states = result_set.excited_states,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_ES_result"),
    pytest.lazy_fixture("turbomole_ES_result"),
    pytest.lazy_fixture("orca_ES_result")
])
def test_pl_diagram(result_set, tmp_path, digichem_options):
    """Can we make an emission spectrum?"""
    maker = Emission_graph_maker.from_options(
        tmp_path / "tmp.png",
        excited_states = result_set.excited_states,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_opt_result"),
    pytest.lazy_fixture("turbomole_opt_result"),
    pytest.lazy_fixture("orca_opt_freq_result")
])
def test_freq_diagram(result_set, tmp_path, digichem_options):
    """Can we make an IR spectrum?"""
    maker = Frequency_graph_maker.from_options(
        tmp_path / "tmp.png",
        vibrations = result_set.vibrations,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("orca_nmr_result"),
])
def test_nmr_diagram(result_set, tmp_path, digichem_options):
    """Can we make an NMR spectrum?"""
    # Get a suitable spectrometer object.
    spectrometer = result_set.nmr.spectrometer

    maker = NMR_graph_maker.from_options(
        tmp_path / "tmp.png",
        graph = spectrometer.get_graph(1, 1),
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("orca_nmr_result"),
])
def test_focus_nmr_diagram(result_set, tmp_path, digichem_options):
    """Can we make an NMR spectrum?"""
    # Get a suitable spectrometer object.
    spectrometer = result_set.nmr.spectrometer

    maker = NMR_graph_zoom_maker.from_options(
        tmp_path / "tmp.png",
        graph = spectrometer.get_graph(1, 1),
        focus = result_set.atoms.groups[(2, "H")],
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_opt_result"),
    pytest.lazy_fixture("turbomole_opt_result"),
    pytest.lazy_fixture("orca_opt_freq_result")
])
def test_2d_diagram(result_set, tmp_path, digichem_options):
    """Can we make an IR spectrum?"""
    maker = Skeletal_image_maker.from_options(
        tmp_path / "tmp.png",
        atoms = result_set.atoms,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_path", [
    Path(data_directory(), "Pyridine/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz")
], ids = ["Gaussian"])
@pytest.mark.parametrize("maker_cls", [
    Structure_image_maker,
    Orbital_image_maker,
])
def test_3d_image(result_path, maker_cls, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = maker_cls.from_options(
            tmp_path / "tmp.png",
            cube_file = Fchk_to_cube.from_options(
                tmp_path / "tmp.cube",
                fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
                options = digichem_options
            ),
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


# @pytest.mark.parametrize("result_path", [
#     Path(data_directory(), "Pyridine/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz")
# ], ids = ["Gaussian"])
# @pytest.mark.parametrize("density", [
#     "SCF"
# ])

@pytest.mark.parametrize("result_path, density", [
    [Path(data_directory(), "Pyridine/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"), "SCF"]
], ids = ["Gaussian"])
def test_density_image(result_path, density, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = Density_image_maker.from_options(
            tmp_path / "tmp.png",
            cube_file = Fchk_to_density_cube.from_options(
                tmp_path / "tmp.SCF.cube",
                fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
                density_type = density,
                options = digichem_options
            ),
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("result_path, density", [
    [Path(data_directory(), "Pyridine/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"), "SCF"]
], ids = ["Gaussian"])
def test_spin_density_image(result_path, density, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = Density_image_maker.from_options(
            tmp_path / "tmp.png",
            cube_file = Fchk_to_spin_cube.from_options(
                tmp_path / "tmp.Spin.cube",
                fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
                spin_density = density,
                options = digichem_options
            ),
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("result_path", [
    Path(data_directory(), "Pyridine/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz")
], ids = ["Gaussian"])
def test_combined_orbital_image(result_path, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = Combined_orbital_image_maker.from_options(
            tmp_path / "tmp.png",
            HOMO_cube_file = Fchk_to_cube.from_options(
                tmp_path / "tmp.HOMO.cube",
                fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
                orbital = "LUMO",
                options = digichem_options
            ),
            LUMO_cube_file = Fchk_to_cube.from_options(
                tmp_path / "tmp.LUMO.cube",
                fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
                orbital = "LUMO",
                options = digichem_options
            ),
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("result_path", [
    Path(data_directory(), "Benzene Anion/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Gas Phase 6-31G(d,p).tar.gz")
], ids = ["Gaussian"])
def test_unrestricted_orbital_image(result_path, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = Orbital_image_maker.from_options(
            tmp_path / "tmp.png",
            cube_file = Fchk_to_cube.from_options(
                tmp_path / "tmp.cube",
                fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
                cubegen_type = "AMO",
                orbital = "HOMO",
                options = digichem_options
            ),
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("result_path", [
    Path(data_directory(), "Pyridine/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz")
])
def test_pdm_image(result_path, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = Permanent_dipole_image_maker.from_options(
            tmp_path / "tmp.png",
            cube_file = Fchk_to_cube.from_options(
                tmp_path / "tmp.cube",
                fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
                options = digichem_options
            ),
            dipole_moment = result_set.dipole_moment,
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("result_path", [
    Path(data_directory(), "Pyridine/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"),
])
def test_tdm_image(result_path, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = Transition_dipole_image_maker.from_options(
            tmp_path / "tmp.png",
            cube_file = Fchk_to_cube.from_options(
                tmp_path / "tmp.cube",
                fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
                options = digichem_options
            ),
            dipole_moment = result_set.excited_states.find("S(1)").transition_dipole_moment.electric,
            magnetic_dipole_moment = result_set.excited_states.find("S(1)").transition_dipole_moment.magnetic,
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("result_path", [
    Path(data_directory(), "Pyridine/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"),
])
def test_nto_image(result_path, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = Orbital_image_maker.from_options(
            tmp_path / "tmp.png",
            cube_file = Fchk_to_nto_cube.from_options(
                tmp_path / "tmp.cube",
                fchk_file = result_set.metadata.auxiliary_files['fchk_file'],
                options = digichem_options
            ),
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


# General imports.
from pathlib import Path
from timeit import default_timer as timer
import datetime
import warnings

from digichem import misc
from digichem.exception.base import File_maker_exception, Digichem_exception

# Digichem imports.
import digichem.log


class File_maker_ABC():
    """
    ABC for file maker classes.
    
    These classes make file(s) on the fly, they are mostly used for image rendering (and making the prerequisite files for image rendering).
    
    Do not try to init an instance of this class yourself; use a real one below (or elsewhere).
    """
    #####################################
    # To be implemented in sub classes. #
    #####################################
    
    def __init__(self, output):
        """
        :param output:  The filename/path to the new file (this path doesn't need to point to a real file yet; we will use this path to write to).
        """
        self.output = output
        self.file_path = {'file': self.output}
        
        # A flag of whether we have already created our output file.
        self.done_file_creation = False
    
    def get_file(self, name = 'file'):
        """
        Get the path to one of the files that this class represents, converting from our input file and writing to file first if necessary.
        """
        raise NotImplementedError("get_file() is not implemented in this ABC")
    
    def delete(self, lazy = False):
        """
        Delete the file(s) described by this object.
        """
        raise NotImplementedError("delete() is not implemented in this ABC")
    
    def get_parallel(self, name = 'file', cpus = 1):
        """
        Generate the file represented by this object in a parallel context.

        get_parallel() will be called by a higher level function as an argument to ThreadPoolExecutor.map() or similar,
        to generate many files simultaneously. This is useful for slow operations (such as cube generation) that are
        difficult to parallelise individually, but easy to parallelise across multiple files.

        :param name: The file to generate.
        :param cpus: The number of CPUs this operation should use, nearly always 1.
        """
        # This default implementation does nothing.
        pass
    
    #################################
    # Implemented in this sub-class #
    #################################
        
    def safe_get_file(self, name = 'file'):
        """
        Get the path to one of the files that this class represents, converting from our input file and writing to file first if necessary.
        
        The functioning of this method is controlled by the dont_modify & use_existing flags.
        
        You can also use the normal python attribute mechanism (either through getattr() or dot notation) to get these paths.
        
        :raises KeyError: If name is not the name of one of the files this class represents.
        :param name: The name of a file to get. Depends on the paths in self.file_path.
        :return: A pathlib Path object pointing to the file represented by name, or None if no file could be created.
        """
        try:
            return self.get_file(name)
        
        except Exception:
            digichem.log.get_logger().error("Unable to get file '{}'".format(self.file_path[name]), exc_info = True)
            
            return None
    
    def relative_path(self, file_name = 'file', *, output_base):
        """
        Get the relative path to the file (or one of the files) represented by this object.
        
        The path is relative to the directory where we write our file, called the output_base.
        
        :param file_name: The name of one of the files that we represent to get the path of, this can be excluded if we only represent one file.
        :param output_base: Pathlib Path to the base directory to get a relative path to.
        :return: A relative Path object to the file.
        """
        # First get the file's real path (this could trigger rendering).        
        file_path = self.safe_get_file(file_name)
        
        # Return None if that doesn't work (eg, because dont_modify = True and use_existing = False) to match the rest of the class.
        if file_path is None:
            return None
                        
        # And return a relative path (this can still raise exceptions).
        return file_path.relative_to(output_base)

    
    def __str__(self):
        return str(self.safe_get_file())


class Dummy_file_maker(File_maker_ABC):
    """
    A dummy file maker that will always fail to make any files.
    
    This class is useful where it is not possible to construct a real file maker object, where instead an instance of this class can be seamlessly used instead.
    """

    def __init__(self, output, message):
        """
        :param output: The path to the file this class is not creating.
        :param message: The reason why this dummy file maker is being used.
        """
        super().__init__(output)
        self.message = message
        
    def get_file(self, name = 'file'):
        """
        Get the path to one of the files that this class represents, converting from our input file and writing to file first if necessary.
        """
        raise File_maker_exception(self, self.message)
    
    def delete(self, lazy = False):
        """
        Delete the file(s) described by this object.
        """
        return False
    

class File_maker(File_maker_ABC):
    """
    Superclass for classes that automatically create files when needed.
    """
    
    # Text description of our output file type, used for error messages etc. This can be changed by inheriting classes.
    output_file_type = "output"
    
    def __init__(self, output, existing_file = None, dont_modify = False, use_existing = False, full_path_names = False):
        """
        Constructor for File_converter objects.
        
        :param output:  The filename/path to the new file (this path doesn't need to point to a real file yet; we will use this path to write to).
        :param existing_file: An optional existing file of the type we're converting to. If this is given, then no conversion is done. This option exists so the user can specify files they have already converted themselves.
        :param dont_modify: Flag that prevents modifying the file on disk. If True, no new file will be written even if one does not already exist. dont_modify is automatically set to True if existing_file is not None (to prevent over-writing whatever file the user gave us).
        :param use_existing: Flag that modifies how file conversion works. If True, existing files will be preferentially used if available (set to False to force overwriting existing files).
        :param full_path_names: Whether to print full path names in log messages.
        """
        # If we've been given an existing_file explicitly, check it exists.
        if existing_file is not None:
            # Convert to path.
            existing_file = Path(existing_file)
            
            # Force set dont_modify = True so we don't overwrite the input that the user has given us.
            dont_modify = True
            digichem.log.get_logger().debug("Setting dont_modify == True to prevent overwrite of user specified {} file '{}'".format(self.output_file_type, existing_file))
            
            # Check
            if not existing_file.exists():
                warnings.warn("The given {} file '{}' does not appear to exist".format(self.output_file_type, existing_file))
                # Continue anyway.
            
            # Set our output appropriately, and continue as normal.
            output = existing_file
        else:
            # Convert to path.
            output = Path(output)
        
        self.dont_modify = dont_modify
        self.use_existing = use_existing
        
        self.failed_file_creation = False
        
        self.full_path_names = full_path_names
        
        super().__init__(output)
    
    def get_file(self, name = 'file'):
        """
        Get the path to one of the files that this class represents, converting from our input file and writing to file first if necessary.
        
        The functioning of this method is controlled by the dont_modify & use_existing flags.
                
        :raises KeyError: If name is not the name of one of the files this class represents.
        :raises File_maker_exception: If the file could not be created for some reason.
        :param name: The name of a file to get. Depends on the paths in self.file_path.
        :return: A pathlib Path object pointing to the file represented by name.
        """
        # We first need to see if we've already made our files or not.
        # Although we can actually represent multiple files, they are all created at the same time.
        if self.done_file_creation:
            return self.file_path[name]
        else:
            # Our files haven't been created. What happens next depends on the options given to us at __init__().
            if self.use_existing:
                # We've been asked to use existing files if they exist, see if they do.
                # There is obviously a race condition here, but we're not making any guarantee to the calling function that the file really exists (or even what it is), we're only checking to see if we need to create a new file or not.
                if self.file_path[name].exists():
                    # File exists, return its path.
                    #digichem.log.get_logger().debug("Using existing {} file '{}'".format(self.output_file_type, self.file_path[name]))
                    return self.file_path[name]
            
            # There are no files we can use. If we're allowed, write new files.
            if not self.dont_modify:
                # Try and make.
                try:
                    self.make()
                    return self.file_path[name]
                
                except File_maker_exception as e:
                    # Don't raise a new exception if we've already got one as this just pollutes the log.
                    self.failed_file_creation = True
                    raise
                
                except Exception as e:
                    # Know that a KeyError could be thrown here if name is not valid, but I think we'd want that to propagate anyway.
                    self.failed_file_creation = True
                    raise File_maker_exception(self, "Unable to create file") from e
            else:
                raise File_maker_exception(self, "Not creating file because dont_modify is True")
    
    @property
    def creation_message(self):
        """
        A short message that may (depending on log-level) be printed to the user before make_files() is called.
        """
        return "Generating {} file '{}'".format(self.output_file_type, self.output if self.full_path_names else self.output.name)
    
    def check_can_make(self):
        """
        Check whether it is feasible to try and create the files(s) that we represent.
        
        Reasons for making not being possible are varied and are up to the inheriting class, but include eg, a required input (cube, fchk) file not being given.
        
        This method returns nothing, but will raise an File_maker_exception exception if the rendering is not possible.
        """
        if self.failed_file_creation:
            # We've already tried to make this file once before and failed, don't try again.
            raise File_maker_exception(self, "Not creating file due to previous errors")
    
    def make(self):
        """
        Make the file(s) described by this object.
        
        This method is a wrapper, used to call other methods that should be written by the user.
        
        Most inheriting classes won't need to modify this.
        
        This method should not be called directly, and doing so can overwrite files that we've been asked not to overwrite.
        """
        # First check if we can render. This method will throw an exception to stop us if we can't proceed.
        self.check_can_make()

        # Print a debug message (because lots can go wrong next and this step an be quite slow).    
        digichem.log.get_logger().info(self.creation_message)
        
        start_timer = timer()
        
        # Make our parent folder(s).
        try:
            self.output.parent.mkdir(parents = True)
        except FileExistsError:
            # This will happen when the dir already exists, which is fine.
            pass
            
        
        # Call the real workhorse, which is implementation specific.
        self.make_files()
        
        # Set our flag.
        self.done_file_creation = True
        
        # Print timing info.
        self.duration = datetime.timedelta(seconds = timer() - start_timer)
        digichem.log.get_logger().info("File generation duration: {} ({} total seconds)".format(misc.timedelta_to_string(self.duration), self.duration.total_seconds()))
        
    def delete(self, lazy = False):
        """
        Delete the file(s) described by this object.
        
        This method will try to delete all files represented by this object, even if exceptions are raised during deletion of an earlier file.
        If lazy is not True, note that it is the last caught exception (from the last file deletion that went wrong) that is re-raised.
        
        :param lazy: If true, no exceptions are raised if file deletion is not possible.
        :return: True if all files were deleted successfully, False otherwise.
        """
        # We'll keep hold of any exceptions for later. 
        exc = None
        
        # Delete each of our files.
        for file_name, file_path in self.file_path.items():
            try:
                # First check we're allowed to delete the file.
                if self.dont_modify:
                    raise Digichem_exception("Unable to delete file '{}'; dont_modify is True".format(file_path))
                
                # DELETE.
                file_path.unlink()
                
                # Unset the done_file_creation flag, so we'll re-write any files if they are called for again.
                self.done_file_creation = False
            except Exception as e:
                # Hold onto our exception. This might discard a previous exception, but we can only raise one exception anyway (?).
                exc = e
        
        # If we've been asked to report exceptions, do so.
        if not lazy and exc is not None:
            raise exc
        
        # Also return whether anything bad happened.
        return exc is None
        
    def make_files(self):
        """
        Make the files(s) described by this object. Inheriting classes should write their own implementation.
        """
        pass


class File_converter(File_maker):
    """
    Superclass for classes that automatically convert between different file formats.
    """
    
    # Text description of our input file type, used for error messages etc. This can be changed by inheriting classes.
    input_file_type = "input"
    
    
    def __init__(self, output, input_file = None, existing_file = None, dont_modify = False, use_existing = True):
        """
        Constructor for File_converter objects.
        
        :param output:  The filename/path to the new file (this path doesn't need to point to a real file yet; we will use this path to write to).
        :param input_file: The input file that will be converted. Not that while this option is formally optional, conversion will normally fail if no input is available.
        :param existing_file: An optional existing file of the type we're converting to. If this is given, then no conversion is done. This option exists so the user can specify files they have already converted themselves.
        :param dont_modify: Flag that modifies how file conversion works. If True, no new file will be written.
        :param use_existing: Flag that modifies how file conversion works. If True, no new file will be written. If True, existing files will be preferentially used if available (set to False to force overwriting existing files).
        """
        # Call our parent.
        super().__init__(output, existing_file, dont_modify, use_existing)
        
        # And save our input file.
        self.input_file = Path(input_file) if isinstance(input_file, str) else input_file
    
    
    def check_can_make(self):
        """
        Check whether it is feasible to try and create the files(s) that we represent.
        
        Reasons for making not being possible are varied and are up to the inheriting class, but include eg, a required input (cube, fchk) file not being given.
        
        This method returns nothing, but will raise an File_maker_exception exception if the rendering is not possible.
        """
        super().check_can_make()
        
        if self.input_file is None:
            raise File_maker_exception(self, "No {} file is available".format(self.input_file_type))
        try:
            if self.input_file.safe_get_file() is None:
                raise File_maker_exception(self, "No {} file is available".format(self.input_file_type))
        except AttributeError:
            # Input file does not have a safe_get_file() method.
            pass
    
    @property
    def creation_message(self):
        """
        A short message that may (depending on log-level) be printed to the user before make_files() is called.
        """
        return "Converting {} file '{}' to {} file '{}'".format(self.input_file_type, self.input_file if self.full_path_names else Path(str(self.input_file)).name, self.output_file_type, self.output if self.full_path_names else Path(str(self.output)).name)        



"""Code for loading/reading/parsing config files and configurable files."""

import yaml

class Config_parser():
    """
    Class for loading standard digichem options (from a string).
    """
    
    def __init__(self, config_string):
        """
        Constructor for Options loader objects.
        
        :param config_string: A string to parse options from.
        """
        self.config_string = config_string
        self.config_path = None
    
    def parse(self, config_stream):
        """
        """
        # Load with PyYaml.
        try:
            config = yaml.safe_load(config_stream)
            
        except Exception as e:
            # We could ignore errors here but it might not be obvious to the user why their settings have been ignored
            raise Exception("Error parsing settings") from e
            
        # Check the config isn't None (which it will be if the file is empty.)
        try:
            if not isinstance(config, dict):
                raise Exception("Config option '{}' is formatted incorrectly".format(config_stream))
            return self.pre_process(config)
        
        except Exception as e:
            if config is None:
                # Don't panic, the file was just empty.
                return self.pre_process({})
            else:
                # Something else went wrong; panic.
                raise e
    
    def load(self):
        """
        Load the config file.
        """
        return self.parse(self.config_string)
    
    def pre_process(self, config):
        """
        Peform some pre-processing on a just-loaded config.
        
        :param config: Dictionary of config options.
        """
        return config
        # return Config(config, FILE_NAME = self.config_path)


class Config_file_parser(Config_parser):
    """
    Class for loading config files.
    """
    
    def __init__(self, path):
        """
        Constructor for Options loader objects.
        
        :param path: Path to the config file to load.
        """
        self.config_path = path
        
    def load(self, not_exists_ok = False):
        """
        Load the config file.
        
        :param not_exists_ok: If False and our config file does not exist, raise an exception.
        """
        try:
            with open(self.config_path, "rt") as config_stream:
                return self.parse(config_stream)
                    
        except FileNotFoundError:
            if not_exists_ok:
                # No need to panic.
                return self.pre_process({})
            
            else:
                # Panic.
                raise


from matplotlib import ticker
import statistics
import math

from digichem.exception.base import Result_unavailable_error
from digichem.image.graph import OneD_graph_image_maker


class Orbital_diagram_maker(OneD_graph_image_maker):
    """
    Class for making orbital energy diagrams.
    """
        
    def __init__(self, output, orbitals, full_axis_lines = False, y_limits = "auto", limits_fallback = True, **kwargs):
        """
        Constructor for graph makers.
        
        :param output: A path to an output file to write to. The extension of this path is used to determine the format of the file (eg: png, jpeg).
        :param orbitals: A list of orbitals to plot.
        :param full_axis_lines: If False, a boundary line is drawn for the x axis but nowhere else. If True, boundary lines are drawn on all 4 sides of the plotting area.
        :param y_limits: String controlling how the y axis limits are set. Options are 'all' for showing all orbitals, 'auto' for standard auto scaling showing the HOMO, LUMO and 0 point, or 'center' for placing the HOMO-LUMO gap in the center of the y axis. Alternatively, limits_method can be a tuple of (y_min, y_max) which will be used directly as axis limits.
        :param limits_fallback: If True, use simple_limits() if the chosen limits_method fails. If false, an exception is raised.
        """
        super().__init__(output, **kwargs)
        self.orbitals = orbitals
        
        # Set our output width and height.
        self.init_image_width = 4.1
        # We use a wider graph if we are using unrestricted orbitals
        if self.orbitals.spin_type != "none":
            self.init_image_width = 4.9
        self.init_image_height = 5.6
        self.output_dpi = 130
        
        # Axes label (no x label here).
        self.y_label = 'Energy /eV'
        # Y-axis scaling.
        self.inch_per_y = 0.62
        
        # Save options.
        self.full_axis_lines = full_axis_lines
        self.limits_method = y_limits
        self.limits_fallback = limits_fallback
        
    @classmethod
    def from_options(self, output, *, orbitals, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """    
        return self(
            output,
            orbitals = orbitals,
            full_axis_lines = options['orbital_diagram']['full_axis_lines'],
            y_limits = options['orbital_diagram']['y_limits'],
            enable_rendering = options['orbital_diagram']['enable_rendering'],
            **kwargs
        )
        
    def plot_lines(self):
        """
        Plot the lines that make up the body of the graph.
        
        This method is called automatically as part of the make() method.
        :return: Nothing.
        """
        # We'll plot occupied and virtual orbitals separately, with different line styles.
        occupied = [mo.energy for mo in self.orbitals if mo.HOMO_difference <= 0]
        virtual = [mo.energy for mo in self.orbitals if mo.HOMO_difference > 0]
        
        # We can use an event plot which is good for 1D graphs like this.
        self.axes.eventplot(occupied, linestyles = "solid", **self.plot_options)
        self.axes.eventplot(virtual, linestyles = "dashed", **self.plot_options)
        
    def plot_labels(self):
        """
        Plot the labels on our graph that identify the HOMO, LUMO and energy gap.
        
        This method is called automatically as part of the make() method.
        :return: Nothing.
        """
        # First work our what our x coord is (this is the same for all points).
        x_pos = self.plot_options["lineoffsets"]
        start_x_pos = self.plot_options["lineoffsets"] - self.plot_options["linelengths"] /2
        
        # Try just plotting labels for HOMO and LUMO (because we're pretty sure there should be space).
        try:
            HOMO = self.orbitals.get_orbital(HOMO_difference = 0)
        except Exception:
            # Couldn't get HOMO.
            HOMO = None
            
        try:
            LUMO = self.orbitals.get_orbital(HOMO_difference = 1)
        except Exception:
            # Couldn't get LUMO.
            LUMO = None
        
        if HOMO is not None:
            # HOMO text goes above the line.
            self.axes.annotate(
                "{}: {:0.2f} eV".format(HOMO.label, HOMO.energy),
                (x_pos, HOMO.energy),
                textcoords = "offset pixels",
                xytext = (0, 7),
                horizontalalignment = "center"
            )

        if LUMO is not None:    
            # LUMO text goes below the line.
            self.axes.annotate(
                "{}: {:0.2f} eV".format(LUMO.label, LUMO.energy),
                (x_pos, LUMO.energy),
                textcoords = "offset pixels",
                xytext = (0, -7),
                verticalalignment = "top",
                horizontalalignment = "center"
            )
        
        if HOMO is not None and LUMO is not None:
            # Now plot our dE HOMO-LUMO arrow.
            self.axes.annotate(
                "",
                xy = (start_x_pos, HOMO.energy),
                xytext = (start_x_pos, LUMO.energy),
                arrowprops=dict(arrowstyle="<->")
            )
            
            # And then our dE HOMO-LUMO energy.
            self.axes.annotate(
                "ΔE: {:0.2f} eV".format(LUMO.energy - HOMO.energy),
                (x_pos, statistics.median([HOMO.energy, LUMO.energy])),
                verticalalignment = "center",
                horizontalalignment = "center"
            )
            
    def all_limits(self):
        """
        Limit the Y axis so all orbitals are visible.
        """
        self.axes.autoscale(enable = True, axis = 'y')
    
    def center_limits(self):
        """
        Limit the Y axis so the HOMO and LUMO are in the center of the diagram, and the 0 energy point is visible.
        
        :raises Result_unavailable_error: If either the HOMO or LUMO is not available.
        """
        # Get our FMOs.
        HOMO = self.orbitals.get_orbital(HOMO_difference = 0)
        LUMO = self.orbitals.get_orbital(HOMO_difference = 1)
        
        # Calculate the midpoint between the FMOs.
        midpoint = HOMO.energy - ((HOMO.energy - LUMO.energy) /2)
        
        # Work out y_max, which is set depending on the position of the LUMO.
        y_max = max(0, math.ceil(LUMO.energy +1.1))
        
        # Y min is then set so that midpoint is indeed in the midpoint.
        y_min = midpoint - (y_max - midpoint)
        
        #
        y_min = round(y_min)
        
        # Set our axis limits.
        self.axes.set_ylim(y_min, y_max)
        
    def standard_limits(self):
        """
        Limit the Y axis so that the HOMO, LUMO and 0 energy point are all visible.
        
        :raises Result_unavailable_error: If either the HOMO or LUMO is not available.
        """
        # Get our FMOs.
        HOMO = self.orbitals.get_orbital(HOMO_difference = 0)
        LUMO = self.orbitals.get_orbital(HOMO_difference = 1)
        
        # Add on a bit to both HOMO and LUMO so we expand if the orbital is close to the end of the axis.
        y_min = math.floor(HOMO.energy -1.1) if HOMO is not None else -8
        y_max = max(0, math.ceil(LUMO.energy +1.1)) if LUMO is not None else 0
        self.axes.set_ylim(y_min, y_max)
    
    def simple_limits(self, y_min = -8, y_max = 0):
        """
        Limit the Y axis between between two points.
        
        :param y_min: The start of the y axis.
        :param y_max: The end of the y axis.
        """
        self.axes.set_ylim(y_min, y_max)
    
    def adjust_axes(self):
        """
        Adjust the axes of our graph.
        
        This method is called automatically as part of the make() method.
        :return: Nothing.
        """
        super().adjust_axes()
        
        # Orbital diagrams only have the left axis spine to enable easy stacking.
        if not self.full_axis_lines:
            self.axes.spines['top'].set_visible(False)
            self.axes.spines['right'].set_visible(False)
            self.axes.spines['bottom'].set_visible(False)
                
        # Adjust axis
        # Our x axis is always the same.
        self.axes.set_xlim(0, 1.5)
        
        # Our Y axis changes with our given data. If we a HOMO and/or LUMO, we'll stretch appropriately to fit them in.
        # If We're missing one of the FMO (unlikely, but possible), we'll resort to semi-random defaults (for now).
        try:
            if self.limits_method == "all":
                self.all_limits()
            elif self.limits_method == "center":
                self.center_limits()
            elif self.limits_method == "auto":
                self.standard_limits()
            elif isinstance(self.limits_method, tuple) or isinstance(self.limits_method, list):
                self.simple_limits(self.limits_method[0], self.limits_method[1])
            else:
                raise ValueError("Unknown limits method '{}'".format(self.limits_method))
        except Result_unavailable_error:
            # HOMO-LUMO not available.
            if self.limits_fallback:
                self.simple_limits()
            else:
                raise
        
        # Change spacing of tick markers.
        # No tick marks for x axis.
        self.axes.set_xticks([])
        self.axes.yaxis.set_major_locator(ticker.MultipleLocator(1))
                
        # Our y axis scales with data.
        self.figure.tight_layout()
        self.constant_scale(1, self.inch_per_y)
        
        


# General imports.
import subprocess
from subprocess import CalledProcessError
import re
import os
import copy
from pathlib import Path
import warnings
import json

# Digichem imports.
from digichem.exception.base import Digichem_exception
import digichem.log


class Openprattle_converter():
    """
    Provides an interface to oprattle.

    Openprattle provides a library interface, but because of the GPL we cannot use it.
    Calling the oprattle command is fine, however.
    """

    def __init__(self, *, input_file = None, input_file_buffer = None, input_file_path = None, input_file_type = None, executable = "oprattle"):
        """
        Constructor for the OpenBabel converter.
        
        Input files can be specified in one of three ways:
         - As an open file descriptor (input_file and input_file_type)
         - As an in-memory buffer, most probably a string or bytes-like object (input_file_buffer and input_file_type)
         - As a file path (input_file_path and optionally input_file_type)
        
        :param input_file: An open file descriptor in the format given by input_file_type that should be converted.
        :param input_file_buffer: Alternatively, a buffer (unicode string or bytes) in the format given by input_file_type that should be converted.
        :param input_file_path: Alternatively, a path to a file that should be converted.
        :param input_file_type: A shortcode identifying the format of the input file. If not given but input_file_path is given, then this will be determined automatically.
        :param executable: Path or command name to the oprattle executable.
        """
        self.input_file = input_file
        self.input_file_buffer = input_file_buffer if input_file is None else input_file.read()
        self.input_file_path = input_file_path
        self.input_file_type = input_file_type
        self.executable = executable
        
    @property
    def input_name(self):
        """
        A descriptive name of the file we are converting. Works even if converting from memory.
        """
        if self.input_file_path is not None:
            return self.input_file_path
        else:
            return "(file loaded from memory)"
        
    @classmethod
    def get_cls(self, input_file_type):
        """
        This function is deprecated, and does nothing.
        """
        warnings.warn("get_cls() is deprecated, use Openprattle_converter directly instead", DeprecationWarning)
        return self
        
    @classmethod
    def from_file(self, input_file_path, input_file_type = None, **kwargs):
        """
        This function is deprecated, and does nothing.
        """
        warnings.warn("from_file() is deprecated, use the Openprattle_converter constructor directly instead", DeprecationWarning)
        return self.get_cls(None)(input_file_path = input_file_path, input_file_type = input_file_type, **kwargs)
    
    @classmethod
    def type_from_file_name(self, input_file_name, allow_none = False):
        """
        Get the type of a file based on its file name.
        
        This method largely uses the file extension (.com, .tmol etc), with a few other simple rules.
        
        :param input_file_name: The name of the file to check.
        :param allow_none: If the type of the file cannot be determined and allow_none is False (the default), an exception is raised. Otherwise, None is returned.
        """
        try:
            input_file_name = Path(input_file_name)
        
        except TypeError:
            if allow_none:
                return None
            
            else:
                # No file given.
                raise ValueError("Could not automatically determine file format; no file name was given") from None
        
        # Get file extension (removing the dot character).
        extension = input_file_name.suffix[1:]
        
        if extension != "":
            # All done.
            return extension.lower()
        
        elif input_file_name.name == "coord":
            # This is a turbomole file.
            return "tmol"
        
        else:
            if allow_none:
                return None
            
            else:
                # Don't recognise the file format.
                raise ValueError("Could not determine file format of file '{}'; the file does not have an extension and is not recognised".format(input_file_name))
        
    def convert(self, output_file_type = None, output_file = None, *, gen3D = None, charge = None, multiplicity = None):
        """
        Convert the input file wrapped by this class to the designated output_file_type.
         
        :param output_file_type: The file type to convert to.
        :param output_file: Optional file name to write to. If not given, the converted file will be returned as a string (or binary string depending on format).
        :param gen3D: If True and the loaded molecule does not have 3D coordinates, these will be generated (this will scramble atom coordinates).        
        :param charge:  Optional charge of the output format.
        :param multiplicity: Optional multiplicity of the output format.
        :return: The converted file, or None if output_file is not None.
        """
        output_file = str(output_file) if output_file is not None else None
        
        # The signature we'll use to run oprattle.
        sig = [
            str(self.executable),
            "--json"
        ]
        
        # Add the input path if we're reading from file.
        if self.input_file_buffer is None:
            sig.append(str(self.input_file_path))
            
        # Now add the input and output switches.
        if output_file_type:
            sig.extend([
                "-o", output_file_type
            ])
        if self.input_file_type:
            sig.extend([
                "-i", self.input_file_type
            ])
        
        # Add gen3D command if we've been asked to.
        if gen3D is True:
            sig.extend(["--gen3D", "True"])
        elif gen3D is False:
            sig.extend(["--gen3D", "False"])
            
        # If a file to write to has been given, set it.
        if output_file is not None:
            sig.extend(['-O', output_file])
        
        # Give our input_file as stdin if we're not reading from file.
        inputs = self.input_file_buffer

        # Encode strings.
        if isinstance(inputs, str):
            inputs = inputs.encode()
        
        # GO.
        done_process = subprocess.run(
            sig,
            input = inputs,
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE,
        )

        # This can throw exceptions.
        self.handle_logging(done_process.stderr.decode())

        if done_process.returncode != 0:
            raise Digichem_exception("prattle subprocess returned code {}".format(done_process.returncode))
        
        # Return our output.
        return done_process.stdout.decode() if output_file is None else None
    
    def handle_logging(self, raw_output):
        """
        """
        exceptions = []
        for raw_message in raw_output.split("\n"):
            if raw_message == "":
                # Nothing returned, nothing to do.
                continue
            
            # Each message should be in JSON, but check.
            try:
                message = json.loads(raw_message)
                message_text = message['message']
                if message['exception']:
                    exceptions.append(Exception(message['exception']))
                    continue
                    #message_text += "\n" + message['exception']
                digichem.log.get_logger().log(
                    message['levelno'],
                    message_text
                )

            except Exception:
                digichem.log.get_logger().error("Unexpected output from oprattle: '{}'".format(raw_message), exc_info=False)

        if len(exceptions) > 0:
            raise exceptions[0]
    


class Oprattle_formats():
    """
    Class for retrieving the supported file formats from oprattle.
    """
    
    def __init__(self, executable = "oprattle"):
        """
        """
        self.executable = executable
    
    def run(self, readwrite):
        """
        """
        if readwrite not in ["read", "write"]:
            raise ValueError("readwrite must be one of either 'read' or 'write'")
        
        # The signature we'll use to run oprattle.
        sig = [
            self.executable,
            "--json",
            "--readable" if readwrite == "read" else "--writable"
        ]
        
        # GO.
        done_process = subprocess.run(
             sig,
             stdout = subprocess.PIPE,
             stderr = subprocess.PIPE,
             universal_newlines = True,
             check = True,
         )
        
        formats = json.loads(done_process.stdout)
        
        return formats

    
    def read(self, refresh = False):
        """
        Retrieve supported input (read) formats.
        """
        if refresh:
            try:
                del self._read
            
            except AttributeError:
                pass
                
        try:
            return self._read
        
        except AttributeError:
            # Cache miss.
            self._read = self.run("read")
            return self._read
        
    def write(self, refresh = False):
        """
        Retrieve supported input (read) formats.
        """
        if refresh:
            try:
                del self._write
            
            except AttributeError:
                pass
        
        try:
            return self._write
        
        except AttributeError:
            # Cache miss.
            self._write = self.run("write")
            return self._write
    
    


from pathlib import Path

class Input_file():
    """
    ABC for classes that support various input format types.
    
    In most cases the concrete class digichem.input.Digichem_coords is what you are likely looking for.
    """
    
    def __init__(self, charge = None, multiplicity = None, name = None, file_name = None, history = None):
        """
        Abstract constructor for Input_file classes.
        
        :param charge: An explicit molecular charge (as an integer). If not given, a charge may be interpreted from other properties of the input file.
        :param multiplicity: An explicit molecular multiplicity (as an integer). If not given, a multiplicity may be interpreted from other properties of the input file.
        :param name: An explicit molecule name. If not given, a name may be interpreted from other properties of the input file.
        :param file_name: The path to the file from which this object was created. If this Input_file was not created from a file (eg, from memory instead), this should be None.
        :param history: The SHA of the previous calculation (if applicable) from which these coordinates were taken.
        """
        if charge is not None and not isinstance(charge, int):
            raise TypeError("Charge must be an integer (or None)")
        if multiplicity is not None and not isinstance(multiplicity, int):
            raise TypeError("Multiplicity must be an integer (or None)")
        
        self.charge = charge
        self.multiplicity = multiplicity
        self.name = name
        # TOOD: This should be called file_path
        self.file_name = Path(file_name) if file_name is not None else None
        self.history = history

    def dump(self):
        """
        Get this input file as a dict.
        """
        return {
            'name': self.name,
            'charge': self.charge,
            'multiplicity': self.multiplicity,
            'history': self.history,
        }
    
    @property
    def implicit_name(self):
        """
        A more intelligent name for this molecule, taking into account our old file name if necessary.
        """
        # If a real name wasn't given, but a file name was, use it.
        if self.name is None and self.file_name is not None:
            name = self.file_name.with_suffix("").name
            
        elif self.name is not None:
            name = self.name
        
        else:
            name = "molecule"
            
        return name

    @property
    def implicit_charge(self):
        """
        The charge of the molecule/system, accounting for cases where no explicit charge is set.
        """ 
        if self.charge is None:
            return 0
        else:
            return self.charge
        
    @property
    def implicit_multiplicity(self):
        """
        The multiplicity (as an integer) of the molecule/system, accounting for cases where no explicit multiplicity is set.
        """ 
        if self.multiplicity is None:
            return 1
        else:
            return self.multiplicity


# General imports.
import math

from digichem.exception import Result_unavailable_error
from digichem.result import Result_container
from digichem.result import Result_object
from digichem.result.base import Floatable_mixin


class SOC_list(Result_container):
    """
    An augmented list containing spin orbit coupling.
    """
    
    def find(self, criteria):
        """
        """
        states = criteria.split(",")
        
        # Check we have two states.
        if len(states) != 2:
            raise ValueError("SOC can only be found between exactly 2 states, not {} states.".format(len(states)))
        
        return self.between(*states)
    
    def between(self, state1, state2, **kwargs):
        """
        Return the spin-orbit coupling object between the two states with given symbols.
        
        :param states: Symbols identifying the two states to get SOC between (eg, "S(1)" "T(1)").
        :param default: A default argument that will be returned if the given states cannot be found.
        """
        states = set((state1, state2))
        
        # Iterate through all our SOC.
        for soc in self:
            # Check if the two states match the two we've been asked to find.
            if len(states.intersection((soc.singlet_state.state_symbol, soc.triplet_state.state_symbol))) == 2:
                # A match!
                return soc
            
        # No match.
        # Return the default if given, otherwise panic.
        if "default" in kwargs:
            return kwargs['default']
        else:
            raise Result_unavailable_error("Spin-orbit coupling", "could not find SOC between states with symbols '{}' and '{}'".format(state1, state2))
    
    @classmethod
    def from_parser(self, parser):
        """
        Get a SOC_list object from an output file parser.
        
        :param parser: An output file parser.
        :return: A SOC_list object. The list will be empty if no SOC is available.
        """
        # Decide which SOC class to use.
        if hasattr(parser.data, "socelements"):
            soc_cls = Spin_orbit_coupling
        else:
            soc_cls = Total_spin_orbit_coupling
        
        return self(soc_cls.list_from_parser(parser))

    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self([(Spin_orbit_coupling if 'soc' in soc_dict else Total_spin_orbit_coupling).from_dump(soc_dict, result_set, options) for soc_dict in data])
    
    def sort(self, *, key = None, **kwargs):
        """
        Sort this list in place.
        """
        # If no key is given, we use a custom default key.
        if key is None:
            key = lambda SOC: (SOC.singlet_state.level, SOC.triplet_state.level)
            
        return super().sort(key = key, **kwargs)


class Spin_orbit_coupling(Result_object, Floatable_mixin):
    """
    Class that represents spin-orbit coupling between two states.
    """
    
    def __init__(self, singlet_state, triplet_state, positive_one, zero, negative_one):
        """
        Constructor for SOC objects.
        
        :param singlet_state: The singlet excited state this coupling is between.
        :param triplet_state: The triplet excited state this coupling is between.
        :param positive_one: SOC between the singlet state and triplet sub-state with quantum number +1 in cm-1.
        :param zero: SOC between the singlet state and triplet sub-state with quantum number 0 in cm-1.
        :param negative_one: SOC between the singlet state and triplet sub-state with quantum number -1 in cm-1.
        """
        self.singlet_state = singlet_state
        self.triplet_state = triplet_state
        self.positive_one = positive_one
        self.zero = zero
        self.negative_one = negative_one
        
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "singlet": self.singlet_state.state_symbol,
            "triplet": self.triplet_state.state_symbol,
            "soc": {
                "+1": {
                    "value": self.positive_one,
                    "units": "c m^-1",
                },
                "0": {
                    "value": self.zero,
                    "units": "c m^-1",
                },
                "-1": {
                    "value": self.negative_one,
                    "units": "c m^-1",
                },
            },
            "rss": {
                "units": "c m^-1",
                "value": self.root_sum_square,
            },
            "mixing_coefficient": self.mixing_coefficient
        }
        
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self(result_set.energy_states.find(data['singlet']), result_set.energy_states.find(data['triplet']), data['soc']['+1']['value'], data['soc']['0']['value'], data['soc']['-1']['value'])
        
    @classmethod
    def list_from_parser(self, parser):
        """
        Create a SOC object from an output file parser.
        """
        # Go through our data.
        try:
            return [self(
                singlet_state = parser.results.energy_states.get_state(state_symbol = singlet_symbol),
                triplet_state = parser.results.energy_states.get_state(state_symbol = triplet_symbol),
                positive_one = positive_soc,
                zero = zero_soc,
                negative_one = negative_soc
            ) for (singlet_symbol, triplet_symbol), (positive_soc, zero_soc, negative_soc) in zip(parser.data.socstates, parser.data.socelements)]
        except AttributeError:
            # No data.
            return []
        
    @property
    def splitting_energy(self):
        """
        The difference in energy between these two states.
        """
        return max(float(self.singlet_state), float(self.triplet_state)) - min(float(self.singlet_state), float(self.triplet_state))
    
    @property
    def mixing_coefficient(self):
        """
        The first order mixing coefficient (λ), given by Hso / dEst.
        Be aware that λ will == math.inf if dEst == 0.
        """
        try:
            return self.energy / self.splitting_energy
        except (FloatingPointError, ZeroDivisionError):
            return math.inf
    
    @property
    def root_sum_square(self):
        """
        The root sum square of this SOC.
        """
        return math.sqrt(self.positive_one **2 + self.zero **2 + self.negative_one **2)
        
    def __float__(self):
        """
        Floatify this SOC class.
        """
        return float(self.root_sum_square)
    
    @property
    def wavenumbers(self):
        """
        SOC in wavenumbers.
        """
        return self.root_sum_square
    
    @property
    def energy(self):
        """
        SOC in eV.
        """
        try:
            return self.wavenumbers_to_energy(self.root_sum_square)
        except (FloatingPointError, ZeroDivisionError):
            return 0.0
    
    def __str__(self):
        """
        Stringify this SOC class.
        """
        return "<{}|Hso|{}> (RSS, +1, 0, -1) (cm-1): {:10.5f} {:10.5f} {:10.5f} {:10.5f}".format(self.singlet_state.state_symbol, self.triplet_state.state_symbol, self.root_sum_square, self.positive_one, self.zero, self.negative_one)


class Total_spin_orbit_coupling(Spin_orbit_coupling):
    """
    A class for representing SOC when only the total is known (and not the individual elements).
    """
    
    def __init__(self, singlet_state, triplet_state, total):
        super().__init__(singlet_state, triplet_state, None, None, None)
        
        self._root_sum_square = total
        
    @property
    def root_sum_square(self):
        return self._root_sum_square
        
    def __str__(self):
        """
        Stringify this SOC class.
        """
        return "<{}|Hso|{}> (cm-1): {:10.5f}".format(self.singlet_state.state_symbol, self.triplet_state.state_symbol, self.root_sum_square)
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        dump_dic = super()._dump_(digichem_options, all)
        # Remove elements.
        del(dump_dic['soc'])
        return dump_dic
    
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self(result_set.energy_states.find(data['singlet']), result_set.energy_states.find(data['triplet']), data['rss']['value'])
        
    @classmethod
    def list_from_parser(self, parser):
        """
        Create a SOC object from an output file parser.
        """
        # Go through our data.
        try:
            return [self(
                singlet_state = parser.results.energy_states.get_state(state_symbol = singlet_symbol),
                triplet_state = parser.results.energy_states.get_state(state_symbol = triplet_symbol),
                total = total
            ) for (singlet_symbol, triplet_symbol), total in zip(parser.data.socstates, parser.data.socenergies)]
        except (AttributeError, Result_unavailable_error):
            # No data.
            return []
        


# General imports.
from PIL import Image
from io import BytesIO

import rdkit.RDLogger
from rdkit.Chem.Draw import rdMolDraw2D

# Digichem imports.
from digichem.image.base import Image_maker


class Skeletal_image_maker(Image_maker):
    """
    A class for rendering skeletal-style molecule structure images.
    """
    
    def __init__(self, output, atoms, *args, abs_resolution = None, rel_resolution = 100, numbering = "group", numbering_font_size = 0.7, explicit_h = False, **kwargs):
        """
        Constructor for Structure_image_maker objects.
        
        :param output: A path to an output file to write to. The extension of this path is used to determine the format of the file (eg: png, jpeg).
        :param atoms: A list of atom of the molecule/system to render.
        :param resolution: The width and height of the rendered image.
        :param render_backend: The library to prefer to use for rendering, possible options are 'rdkit' or 'obabel'. If 'rdkit' is chosen but is not available, obabel will be used as a fallback. 
        :param numbering: Whether to show atom numberings, either False (off), "atomic" (for atom wise) or "group" (for group wise).
        :param explicit_h: Whether to show explicit H.
        """
        self.atoms = atoms
        self.abs_resolution = abs_resolution
        self.rel_resolution = rel_resolution
        self.numbering = numbering
        self.numbering_font_size = numbering_font_size
        self.explicit_h = explicit_h
        super().__init__(output, *args, **kwargs)
        
    @classmethod
    def from_options(self, output, *, atoms, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """    
        return self(
            output,
            atoms = atoms,
            abs_resolution = options['skeletal_image']['resolution']['absolute'],
            rel_resolution = options['skeletal_image']['resolution']['relative'],
            #render_backend = options['skeletal_image']['render_backend'],
            **kwargs
        )
        
    def make_files(self):
        """
        Make the image described by this object.
        """
        if self.abs_resolution:
            resolution = self.abs_resolution
        
        else:
            resolution = int(self.rel_resolution * self.atoms.X_length)
        
        molecule = self.atoms.to_rdkit_molecule()
        
        # Calculate atom groupings.
        atom_groups = self.atoms.groups
        
        for atom in molecule.GetAtoms():
            
            group = [group for group in atom_groups.values() if atom.GetIdx() +1 in [atom.index for atom in group.atoms]][0]
        
            if self.numbering == "both":
                # Add atom labelling.
                #atom.SetProp("molAtomMapNumber", str(atom.GetIdx()+1))
                atom.SetProp("atomNote",  "{} ({})".format(group.id[0], atom.GetIdx()+1))
            
            elif self.numbering == "atomic":
                atom.SetProp("atomNote",  "{}".format(atom.GetIdx()+1))
            
            elif self.numbering == "group":
                atom.SetProp("atomNote",  "{}".format(group.id[0]))
        
        # Remove C-H, if we've been asked to.
        if not self.explicit_h:
            edit_mol = rdkit.Chem.EditableMol(molecule)
            atoms = list(edit_mol.GetMol().GetAtoms())
            for atom_index, atom in enumerate(reversed(atoms)):
                # If this atom is a hydrogen, and it has a single bond to a carbon, delete it.
                if atom.GetSymbol() == "H":
                    bonds = list(atom.GetBonds())
                    if len(bonds) == 1 and bonds[0].GetOtherAtom(atom).GetSymbol() == "C":
                        # This is an implicit H.
                        edit_mol.RemoveAtom(atom.GetIdx())
                        
            molecule = edit_mol.GetMol()
        
        # Different libraries for generating 2D depictions.
        # Coordgen is supposedly superior.
        #rdkit.Chem.AllChem.Compute2DCoords(molecule)
        
        ps = rdkit.Chem.rdCoordGen.CoordGenParams()
        ps.minimizerPrecision = ps.sketcherBestPrecision
        rdkit.Chem.rdCoordGen.AddCoords(molecule, ps)
        rdkit.Chem.rdDepictor.NormalizeDepiction(molecule)
        
        # Then write the file, setting any options we need to.
        d = rdMolDraw2D.MolDraw2DCairo(resolution, resolution)
        d.drawOptions().annotationFontScale = self.numbering_font_size
        d.DrawMolecule(molecule)
        d.FinishDrawing()
        # To save directly, but we're going to open in PIL to crop.
        #d.WriteDrawingText(str(self.output))
        
        # Crop it to remove whitespace.
        with Image.open(BytesIO(d.GetDrawingText()), "r") as im:
            cropped_image = self.auto_crop_image(im)
        
        cropped_image.save(self.output)



import numpy
import os
from pathlib import Path

# Set numpy errors (not sure why this isn't the default...)
numpy.seterr(invalid = 'raise', divide = 'raise')

# Expand path to include mocks.
os.environ["PATH"] = str(Path(__file__).parent / "mock") + os.pathsep + os.environ["PATH"]


from PIL import Image
import numpy as np

from digichem.file import File_maker
from digichem.exception import File_maker_exception

class Cropable_mixin():
    """Mixin class for those that can crop their images."""
    
    @classmethod
    def auto_crop_image(self, im):
        """
        Automatically remove excess white pixels around the outside of our image.
        
        :param im: A PIL Image object.
        :returns: A new PIL Image object with white-space removed.
        """
        # (try to) load our image from file.
        #im = Image.open(file_path, "r")
        im.load()
        
        # This solution was inspired by https://stackoverflow.com/questions/14211340/automatically-cropping-an-image-with-python-pil
        image_data = np.asarray(im)
        if im.mode == "RGBA":
            image_data_bw = image_data[:,:,3]

            # Now we just check which rows and columns are not pure white.
            non_empty_columns = np.where(image_data_bw.max(axis = 0) != 0)[0]
            non_empty_rows = np.where(image_data_bw.max(axis = 1) != 0)[0]
        
        else:
            # Axis 2 is the tuple/array of RGB values for each pixel. We want to know which ones are fully white (255,255,255), so we'll take the min of the 3 values and see if this is equal to 255. If it is, then all the pixels are 255        
            image_data_bw = image_data.min(axis = 2)
        
            # Now we just check which rows and columns are not pure white.
            non_empty_columns = np.where(image_data_bw.min(axis = 0) != 255)[0]
            non_empty_rows = np.where(image_data_bw.min(axis = 1) != 255)[0]

        if len(non_empty_rows) == 0 or len(non_empty_columns) == 0:
            raise ValueError("The image is empty!")

        # Get the bounding box of non-white stuff.
        cropBox = (min(non_empty_rows), max(non_empty_rows), min(non_empty_columns), max(non_empty_columns))
        
        # Copy the image data we want.
        image_data_new = image_data[cropBox[0]:cropBox[1]+1, cropBox[2]:cropBox[3]+1 , :]
        
        # And save over our old file.
        new_image = Image.fromarray(image_data_new)
        return new_image


class Image_maker(File_maker, Cropable_mixin):
    """
    Top level class for image maker objects.
    """
    
    # Text description of our output file type, used for error messages etc. This can be changed by inheriting classes.
    output_file_type = "image"
    
    def __init__(self, *args, enable_rendering = True, **kwargs):
        """
        General constructor for Image_maker objects.
        
        These object are used to represent and make images. Note that a single Image_maker object can represent several image_files
        
        :param output: A path to an output file to write to. How exactly this operates depends on the inheriting class, it is often only used as a base file name. See the class you are using.
        :param dont_modify: Flag that modifies how image creation works. If True, no new images will be written to file.
        :param use_existing: Flag that modifies how image creation works. If True, existing files will be preferentially used if available (set to False to force overwriting existing files).
        """
        super().__init__(*args, existing_file = None, dont_modify = not enable_rendering, **kwargs)
        
    def get_image(self, name = 'file'):
        """
        Get the path to one of the images that this class represents, rendering the image to file first if necessary.
        
        The functioning of this method is controlled by the dont_modify & use_existing flags.
        
        You can also use the normal python attribute mechanism (either through getattr() or dot notation) to get these paths.
        
        :raises KeyError: If name is not the name of one of the images this class represents.
        :param name: The name of an image to get. Depends on the images in self.file_path.
        :return: A pathlib Path object pointing to the image represented by name, or None if no image could be created.
        """
        return self.safe_get_file(name)
    
    @property
    def creation_message(self):
        """
        A short message that may (depending on log-level) be printed to the user before make_files() is called.
        """
        return "Rendering '{}' to file(s)".format(self.output if self.full_path_names else self.output.name)
    
        
    def get_constrained_size(self, max_width, max_height):
        """
        Get the maximum possible dimensions of this image that retain the same aspect ratio that don't exceed a set of dimensions.
        
        :param max_width: The maximum width to resize to; set to math.inf for no max.
        :param max_heigh: The maximum height to resize to; set to math.inf for no max.
        """
        # Weasyprint hack for broken max-width.
                
        # Now we open our diagram with pillow.
        im = Image.open(self.get_image())
        
        width, height = im.size
        
        scale_factor = min(max_width/width, max_height/height)
        
        return (width *scale_factor, height * scale_factor)
    
        